from typing import Any, Dict, List

import numpy as np


def decode_detections(
    input_size: float,
    stride: int,
    score_threshold: float,
    cls: np.ndarray,
    obj: np.ndarray,
    bbox: np.ndarray,
    kps: np.ndarray,
) -> List[Dict[str, Any]]:
    """Decode the detections from neural network output tensors.

    @param input_size: The input size of the model that produced the detections, (width, height).
    @type input_size: float
    @param stride: The stride used in the detection grid.
    @type stride: int
    @param rows: Number of rows in the detection grid.
    @type rows: int
    @param cols: Number of columns in the detection grid.
    @type cols: int
    @param score_threshold: Minimum score threshold for a detection to be considered valid.
    @type score_threshold: float
    @param cls: 2D array of class scores for each grid cell, shape (grid_size, num_classes).
    @type cls: np.ndarray
    @param obj: 1D array of objectness scores for each grid cell, shape (grid_size,).
    @type obj: np.ndarray
    @param bbox: 2D array of bounding box coordinates, shape (grid_size, 4).
    @type bbox: np.ndarray
    @param kps: 2D array of keypoint coordinates, shape (grid_size, num_keypoints * 2).
    @type kps: np.ndarray

    @return: A list of detections, where each detection is a dictionary containing:
        - "bbox": [x1, y1, width, height] (relative bounding box coordinates)
        - "label": int (class label)
        - "keypoints": List[float] (relative keypoint coordinates)
        - "score": float (detection score)
    @rtype: List[Dict[str, Any]]
    """

    input_width, input_height = input_size
    cols = int(input_size[0] / stride)  # w/stride
    rows = int(input_size[1] / stride)  # h/stride

    # Compute the indices
    r, c = np.meshgrid(np.arange(rows), np.arange(cols), indexing="ij")
    idx = r * cols + c

    # Decode scores
    cls_scores = np.clip(cls[idx], 0, 1)
    obj_scores = np.clip(obj[idx], 0, 1)
    max_cls_scores = np.max(cls_scores, axis=-1)
    scores = np.sqrt(max_cls_scores * obj_scores)

    # Get the labels with the highest score
    labels = np.argmax(cls_scores, axis=-1)

    # Decode bounding boxes
    cx = (c + bbox[idx, 0]) * stride
    cy = (r + bbox[idx, 1]) * stride
    w = np.exp(bbox[idx, 2]) * stride
    h = np.exp(bbox[idx, 3]) * stride
    x1 = cx - w / 2
    y1 = cy - h / 2

    # Decode keypoints
    lx = (kps[idx, ::2] + c[:, :, None]) * stride
    ly = (kps[idx, 1::2] + r[:, :, None]) * stride
    keypoints = np.stack((lx / input_width, ly / input_height), axis=-1)

    # Filter detections based on score_threshold
    mask = scores > score_threshold

    # Boolean indexing
    mask_indices = np.where(mask)
    x1_filtered = x1[mask_indices] / input_width
    y1_filtered = y1[mask_indices] / input_height
    w_filtered = w[mask_indices] / input_width
    h_filtered = h[mask_indices] / input_height
    labels_filtered = labels[mask_indices]
    keypoints_filtered = keypoints[mask_indices]
    scores_filtered = scores[mask_indices]

    # Construct the list of dictionaries
    detections = [
        {
            "bbox": [x1, y1, w, h],
            "label": int(label),
            "keypoints": [
                (x, y) for x, y in keypoints
            ],  # convert keypoints to list of tuples
            "score": float(score),
        }
        for x1, y1, w, h, label, keypoints, score in zip(
            x1_filtered,
            y1_filtered,
            w_filtered,
            h_filtered,
            labels_filtered,
            keypoints_filtered,
            scores_filtered,
        )
    ]

    return detections



import depthai as dai

from .visualizers import (
    visualize_classification,
    visualize_detections,
    visualize_detections_xyxy,
    visualize_fastsam,
    visualize_image,
    visualize_keypoints,
    visualize_lane_detections,
    visualize_line_detections,
    visualize_map,
    visualize_segmentation,
    visualize_text_recognition,
    visualize_yolo_extended,
)

visualizers_dict = {
    "YuNetParser": visualize_detections,
    "SCRFDParser": visualize_detections,
    "MPPalmDetectionParser": visualize_detections,
    "YOLO": visualize_detections_xyxy,
    "SSD": visualize_detections,
    "SegmentationParser": visualize_segmentation,
    "MLSDParser": visualize_line_detections,
    "KeypointParser": visualize_keypoints,
    "HRNetParser": visualize_keypoints,
    "SuperAnimalParser": visualize_keypoints,
    "MPHandLandmarkParser": visualize_keypoints,
    "ClassificationParser": visualize_classification,
    "ImageOutputParser": visualize_image,
    "MapOutputParser": visualize_map,
    "YOLOExtendedParser": visualize_yolo_extended,
    "LaneDetectionParser": visualize_lane_detections,
    "FastSAMParser": visualize_fastsam,
    "PPTextDetectionParser": visualize_detections,
    "ClassificationSequenceParser": visualize_text_recognition,
}


def visualize(
    frame: dai.ImgFrame, message: dai.Buffer, parser_name: str, extraParams: dict
):
    """Calls the appropriate visualizer based on the parser name and returns True if the
    pipeline should be stopped."""
    visualizer = visualizers_dict[parser_name]
    return visualizer(frame, message, extraParams)



from .base_parser import BaseParser


class DetectionParser(BaseParser):
    """Parser class for parsing the output of a detection model. The parser expects the
    output of the model to be in the (x_min, y_min, x_max, y_max, confidence) format. As
    the result, the node sends out the detected objects in the form of a message
    containing bounding boxes and confidence scores.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    conf_threshold : float
        Confidence score threshold of detected bounding boxes.
    iou_threshold : float
        Non-maximum suppression threshold.
    max_det : int
        Maximum number of detections to keep.

    Output Message/s
        -------
        **Type**: ImgDetectionsExtended

        **Description**: ImgDetectionsExtended message containing bounding boxes and confidence scores of detected objects.
    ----------------
    """

    def __init__(
        self,
        conf_threshold: float,
        iou_threshold: float,
        max_det: int,
    ) -> None:
        """Initializes the parser node.

        @param conf_threshold: Confidence score threshold of detected bounding boxes.
        @type conf_threshold: float
        @param iou_threshold: Non-maximum suppression threshold.
        @type iou_threshold: float
        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        """
        super().__init__()
        self.conf_threshold = conf_threshold
        self.iou_threshold = iou_threshold
        self.max_det = max_det

    def setConfidenceThreshold(self, threshold: float) -> None:
        """Sets the confidence score threshold for detected objects.

        @param threshold: Confidence score threshold for detected objects.
        @type threshold: float
        """
        if not isinstance(threshold, float):
            raise ValueError("Confidence threshold must be a float.")
        self.conf_threshold = threshold

    def setIOUThreshold(self, threshold: float) -> None:
        """Sets the non-maximum suppression threshold.

        @param threshold: Non-maximum suppression threshold.
        @type threshold: float
        """
        if not isinstance(threshold, float):
            raise ValueError("IOU threshold must be a float.")
        self.iou_threshold = threshold

    def setMaxDetections(self, max_det: int) -> None:
        """Sets the maximum number of detections to keep.

        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        """
        if not isinstance(max_det, int):
            raise ValueError("Max detections must be an integer.")
        self.max_det = max_det

    def build(self, head_config) -> "DetectionParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: DetectionParser
        """

        self.conf_threshold = head_config.get("conf_threshold", self.conf_threshold)
        self.iou_threshold = head_config.get("iou_threshold", self.iou_threshold)
        self.max_det = head_config.get("max_det", self.max_det)

        return self



from typing import List

import depthai as dai
import numpy as np
from numpy.typing import NDArray

from .keypoints import Keypoint
from .segmentation import SegmentationMask


class ImgDetectionExtended(dai.Buffer):
    """A class for storing image detections in (x_center, y_center, width, height)
    format with additional angle and keypoints.

    Attributes
    ----------
    x_center: float
        The X coordinate of the center of the bounding box, relative to the input width.
    y_center: float
        The Y coordinate of the center of the bounding box, relative to the input height.
    width: float
        The width of the bounding box, relative to the input width.
    height: float
        The height of the bounding box, relative to the input height.
    angle: float
        The angle of the bounding box expressed in degrees.
    confidence: float
        Confidence of the detection.
    label: int
        Label of the detection.
    keypoints: List[Keypoint]
        Keypoints of the detection.
    """

    def __init__(self):
        """Initializes the ImgDetectionExtended object."""
        super().__init__()
        self._x_center: float
        self._y_center: float
        self._width: float
        self._height: float

        self._angle: float = 0.0
        self._confidence: float = -1.0
        self._label: int = -1
        self._keypoints: List[Keypoint] = []

    @property
    def x_center(self) -> float:
        """Returns the X coordinate of the center of the bounding box.

        @return: X coordinate of the center of the bounding box.
        @rtype: float
        """
        return self._x_center

    @x_center.setter
    def x_center(self, value: float):
        """Sets the X coordinate of the center of the bounding box.

        @param value: X coordinate of the center of the bounding box.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("X center must be a float.")
        if value < 0 or value > 1:
            raise ValueError("X center must be between 0 and 1.")
        self._x_center = value

    @property
    def y_center(self) -> float:
        """Returns the Y coordinate of the center of the bounding box.

        @return: Y coordinate of the center of the bounding box.
        @rtype: float
        """
        return self._y_center

    @y_center.setter
    def y_center(self, value: float):
        """Sets the Y coordinate of the center of the bounding box.

        @param value: Y coordinate of the center of the bounding box.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("Y center must be a float.")
        if value < 0 or value > 1:
            raise ValueError("Y center must be between 0 and 1.")
        self._y_center = value

    @property
    def width(self) -> float:
        """Returns the width of the bounding box.

        @return: Width of the bounding box.
        @rtype: float
        """
        return self._width

    @width.setter
    def width(self, value: float):
        """Sets the width of the bounding box.

        @param value: Width of the bounding box.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("Width must be a float.")
        if value < 0 or value > 1:
            raise ValueError("Width must be between 0 and 1.")

        self._width = value

    @property
    def height(self) -> float:
        """Returns the height of the bounding box.

        @return: Height of the bounding box.
        @rtype: float
        """
        return self._height

    @height.setter
    def height(self, value: float):
        """Sets the height of the bounding box.

        @param value: Height of the bounding box.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("Height must be a float.")
        if value < 0 or value > 1:
            raise ValueError("Height must be between 0 and 1.")
        self._height = value

    @property
    def angle(self) -> float:
        """Returns the angle of the bounding box.

        @return: Angle of the bounding box.
        @rtype: float
        """
        return self._angle

    @angle.setter
    def angle(self, value: float):
        """Sets the angle of the bounding box.

        @param value: Angle of the bounding box.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise TypeError: If value is not between -360 and 360.
        """
        if not isinstance(value, float):
            raise TypeError("Angle must be a float.")
        if value < -360 or value > 360:
            raise TypeError("Angle must be between -360 and 360 degrees.")
        self._angle = value

    @property
    def confidence(self) -> float:
        """Returns the confidence of the detection.

        @return: Confidence of the detection.
        @rtype: float
        """
        return self._confidence

    @confidence.setter
    def confidence(self, value: float):
        """Sets the confidence of the detection.

        @param value: Confidence of the detection.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("Confidence must be a float.")
        if value < 0 or value > 1:
            raise ValueError("Confidence must be between 0 and 1.")
        self._confidence = value

    @property
    def label(self) -> int:
        """Returns the label of the detection.

        @return: Label of the detection.
        @rtype: int
        """
        return self._label

    @label.setter
    def label(self, value: int):
        """Sets the label of the detection.

        @param value: Label of the detection.
        @type value: int
        @raise TypeError: If value is not an integer.
        """
        if not isinstance(value, int):
            raise TypeError("Label must be an integer.")
        self._label = value

    @property
    def keypoints(
        self,
    ) -> List[Keypoint]:
        """Returns the keypoints.

        @return: List of keypoints.
        @rtype: Keypoints
        """
        return self._keypoints

    @keypoints.setter
    def keypoints(
        self,
        value: List[Keypoint],
    ) -> None:
        """Sets the keypoints.

        @param value: List of keypoints.
        @type value: List[Keypoint]
        @raise TypeError: If value is not a list.
        @raise TypeError: If each element is not of type Keypoint.
        """
        if not isinstance(value, list):
            raise ValueError("Keypoints must be a list")
        if not all(isinstance(item, Keypoint) for item in value):
            raise ValueError("Keypoints must be a list of Keypoint objects.")
        self._keypoints = value


class ImgDetectionsExtended(dai.Buffer):
    """ImgDetectionsExtended class for storing image detections with keypoints.

    Attributes
    ----------
    detections: List[ImgDetectionExtended]
        Image detections with keypoints.
    masks: np.ndarray
        The segmentation masks of the image. All masks are stored in a single numpy array.
    """

    def __init__(self) -> None:
        """Initializes the ImgDetectionsExtended object."""
        super().__init__()
        self._detections: List[ImgDetectionExtended] = []
        self._masks: SegmentationMask = SegmentationMask()

    @property
    def detections(self) -> List[ImgDetectionExtended]:
        """Returns the image detections with keypoints.

        @return: List of image detections with keypoints.
        @rtype: List[ImgDetectionExtended]
        """
        return self._detections

    @detections.setter
    def detections(self, value: List[ImgDetectionExtended]):
        """Sets the image detections with keypoints.

        @param value: List of image detections with keypoints.
        @type value: List[ImgDetectionExtended]
        @raise TypeError: If value is not a list.
        @raise TypeError: If each element is not of type ImgDetectionExtended.
        """
        if not isinstance(value, list):
            raise TypeError("Detections must be a list.")
        if not all(isinstance(detection, ImgDetectionExtended) for detection in value):
            raise TypeError(
                "Detections must be a list of ImgDetectionExtended objects."
            )
        self._detections = value

    @property
    def masks(self) -> NDArray[np.int8]:
        """Returns the segmentation masks stored in a single numpy array.

        @return: Segmentation masks.
        @rtype: SegmentationMask
        """
        return self._masks.mask

    @masks.setter
    def masks(self, value: NDArray[np.int8]):
        """Sets the segmentation mask.

        @param value: Segmentation mask.
        @type value: NDArray[np.int8])
        @raise TypeError: If value is not a numpy array.
        @raise ValueError: If value is not a 2D numpy array.
        @raise ValueError: If each element is not of type int8.
        @raise ValueError: If each element is larger or equal to -1.
        """
        if not isinstance(value, np.ndarray):
            raise TypeError("Mask must be a numpy array.")
        if value.ndim != 2:
            raise ValueError("Mask must be 2D.")
        if value.dtype != np.int8:
            raise ValueError("Mask must be an array of int8.")
        if np.any((value < -1)):
            raise ValueError("Mask must be an array values larger or equal to -1.")
        masks_msg = SegmentationMask()
        masks_msg.mask = value
        self._masks = masks_msg



from typing import List, Union

import depthai as dai
import numpy as np

from ...messages import SegmentationMasksSAM


def create_segmentation_message(x: np.ndarray) -> dai.ImgFrame:
    """Create a DepthAI message for segmentation mask.

    @param x: Segmentation map array of the shape (H, W, E{1}) where E{1} stands for the
        class of the segmented objects.
    @type x: np.array
    @return: Output segmentaion message in ImgFrame.Type.RAW8.
    @rtype: dai.ImgFrame
    @raise ValueError: If the input is not a numpy array.
    @raise ValueError: If the input is not 3D.
    @raise ValueError: If the input 3rd dimension is not E{1}.
    """

    if not isinstance(x, np.ndarray):
        raise ValueError(f"Expected numpy array, got {type(x)}.")
    if len(x.shape) != 3:
        raise ValueError(f"Expected 3D input, got {len(x.shape)}D input.")
    if x.shape[2] != 1:
        raise ValueError(
            f"Expected 1 channel in the third dimension, got {x.shape[2]} channels."
        )
    if isinstance(x[0, 0, 0], float):
        raise ValueError(f"Expected int type, got {type(x[0, 0, 0])}.")
    imgFrame = dai.ImgFrame()
    imgFrame.setFrame(x)
    imgFrame.setWidth(x.shape[1])
    imgFrame.setHeight(x.shape[0])
    imgFrame.setType(dai.ImgFrame.Type.RAW8)
    return imgFrame


def create_sam_message(x: Union[np.ndarray, List[np.ndarray]]) -> SegmentationMasksSAM:
    """Create a DepthAI message for segmentation masks.

    @param x: List of segmentation map arrays of the shape (N, H, W).
    @type x: Union[np.array, List[np.array]]
    @return: Output segmentaion message in SegmentationMasks.
    @rtype: SegmentationMasks
    @raise ValueError: If the input is not a numpy array or list of numpy arrays.
    @raise ValueError: If the input is not 3D.
    """
    if not isinstance(x, (np.ndarray, list)):
        raise ValueError(f"Expected numpy array or list, got {type(x)}.")
    for mask in x:
        if not isinstance(mask, np.ndarray):
            raise ValueError(f"Expected numpy array, got {type(mask)}.")
        if len(mask.shape) != 2:
            raise ValueError(f"Expected 2D input, got {len(mask.shape)}D input.")

    masks_msg = SegmentationMasksSAM()
    if len(x) != 0:
        masks_msg.masks = x if isinstance(x, np.ndarray) else np.array(x)
    return masks_msg



import depthai as dai
import numpy as np

from ...messages import (
    ImgDetectionExtended,
    ImgDetectionsExtended,
    Line,
    Lines,
)
from ...parsers.utils import transform_to_keypoints


def create_detection_message(
    bboxes: np.ndarray,
    scores: np.ndarray,
    angles: np.ndarray = None,
    labels: np.ndarray = None,
    keypoints: np.ndarray = None,
    keypoints_scores: np.ndarray = None,
    masks: np.ndarray = None,
) -> ImgDetectionsExtended:
    """Create a DepthAI message for object detection. The message contains the bounding
    boxes in X_center, Y_center, Width, Height format with optional angles, labels and
    detected object keypoints and masks.

    @param bbox: Bounding boxes of detected objects in (x_center, y_center, width,
        height) format.
    @type bbox: np.ndarray
    @param scores: Confidence scores of the detected objects of shape (N,).
    @type scores: np.ndarray
    @param angles: Angles of detected objects expressed in degrees.
    @type angles: Optional[np.ndarray]
    @param labels: Labels of detected objects of shape (N,).
    @type labels: Optional[np.ndarray]
    @param keypoints: Keypoints of detected objects of shape (N, n_keypoints, dim) where
        dim is 2 or 3.
    @type keypoints: Optional[np.array]
    @param keypoints_scores: Confidence scores of detected keypoints of shape (N,
        n_keypoints, 1).
    @type keypoints_scores: Optional[np.ndarray]
    @param masks: Masks of detected objects of shape (H, W).
    @type masks: Optional[np.ndarray]
    @return: Message containing the bounding boxes, labels, confidence scores, and
        keypoints of detected objects.
    @rtype: ImgDetectionsExtended
    @raise ValueError: If the bboxes are not a numpy array.
    @raise ValueError: If the bboxes are not of shape (N,4).
    @raise ValueError: If the scores are not a numpy array.
    @raise ValueError: If the scores are not of shape (N,).
    @raise ValueError: If the scores do not have the same length as bboxes.
    @raise ValueError: If the angles do not have the same length as bboxes.
    @raise ValueError: If the angles are not between -360 and 360.
    @raise ValueError: If the labels are not a list of integers.
    @raise ValueError: If the labels do not have the same length as bboxes.
    @raise ValueError: If the keypoints are not a numpy array of shape (N, M, 2 or 3).
    @raise ValueError: If the masks are not a 3D numpy array of shape (img_height,
        img_width, N) or (N, img_height, img_width).
    @raise ValueError: If the masks are not in the range [0, 1].
    @raise ValueError: If the keypoints scores are not a numpy array.
    @raise ValueError: If the keypoints scores are not of shape [n_detections,
        n_keypoints, 1].
    @raise ValueError: If the keypoints scores do not have the same length as keypoints.
    @raise ValueError: If the keypoints scores are not between 0 and 1.
    """

    if not isinstance(bboxes, np.ndarray):
        raise ValueError(f"Bounding boxes should be a numpy array, got {type(bboxes)}.")

    if len(bboxes) == 0:
        return ImgDetectionsExtended()

    if len(bboxes.shape) != 2:
        raise ValueError(
            f"Bounding boxes should be a 2D array like [... , [x_center, y_center, width, height], ... ], got {bboxes.shape}."
        )
    if bboxes.shape[1] != 4:
        raise ValueError(
            f"Bounding boxes should be of shape (n, 4), got {bboxes.shape}."
        )

    if not isinstance(scores, np.ndarray):
        raise ValueError(f"Scores should be a numpy array, got {type(scores)}.")

    if len(scores.shape) != 1:
        raise ValueError(f"Scores should be of shape (N,), got {scores.shape}.")

    n_bboxes = len(bboxes)
    if len(scores) != n_bboxes:
        raise ValueError(
            f"Scores should have same length as bboxes, got {len(scores)} scores and {n_bboxes} bounding boxes."
        )

    if labels is not None:
        if not isinstance(labels, np.ndarray):
            raise ValueError(f"Labels should be a numpy array, got {type(labels)}.")
        if labels.shape[0] != n_bboxes:
            raise ValueError(
                f"Labels should have same length as bboxes, got {len(labels)} labels and {n_bboxes} bounding boxes."
            )

    if angles is not None:
        if not isinstance(angles, np.ndarray):
            raise ValueError(f"Angles should be a numpy array, got {type(angles)}.")
        if len(angles) != n_bboxes:
            raise ValueError(
                f"Angles should have same length as bboxes, got {len(angles)} angles and {n_bboxes} bounding boxes."
            )
        for angle in angles:
            if not -360 <= angle <= 360:
                raise ValueError(f"Angles should be between -360 and 360, got {angle}.")

    if keypoints is not None:
        if not isinstance(keypoints, np.ndarray):
            raise ValueError(
                f"Keypoints should be a numpy array, got {type(keypoints)}."
            )
        if len(keypoints.shape) != 3:
            raise ValueError(
                f"Keypoints should be of shape (N, M, 2 or 3) meaning [..., [x, y] or [x, y, z], ...], got {keypoints.shape}."
            )
        n_detections, n_keypoints, dim = keypoints.shape
        keypoints = np.array(keypoints, dtype=float)
        if n_detections != n_bboxes:
            raise ValueError(
                f"Keypoints should have same length as bboxes, got {n_detections} keypoints and {n_bboxes} bounding boxes."
            )
        if dim not in [2, 3]:
            raise ValueError(
                f"Keypoints should be of dimension 2 or 3 e.g. [x, y] or [x, y, z], got {dim} dimensions."
            )

    if keypoints_scores is not None:
        if keypoints is None:
            raise ValueError(
                "Keypoints scores should be provided only if keypoints are provided."
            )
        if not isinstance(keypoints_scores, np.ndarray):
            raise ValueError(
                f"Keypoints scores should be a numpy array, got {type(keypoints_scores)}."
            )

        n_detections, n_keypoints, _ = keypoints.shape
        if keypoints_scores.shape[0] != n_detections:
            raise ValueError(
                f"Keypoints scores should have same length as keypoints, got {len(keypoints_scores)} keypoints scores and {n_detections} keypoints."
            )

        if keypoints_scores.shape[1] != n_keypoints:
            raise ValueError(
                f"Number of keypoints scores per detection should be the same as number of keypoints per detection, got {keypoints_scores.shape[1]} keypoints scores and {n_keypoints} keypoints."
            )

        if not all(0 <= score <= 1 for score in keypoints_scores.flatten()):
            raise ValueError("Keypoints scores should be between 0 and 1.")

    if masks is not None:
        if not isinstance(masks, np.ndarray):
            raise ValueError(f"Masks should be a numpy array, got {type(masks)}.")

        if len(masks.shape) != 2:
            raise ValueError(f"Masks should be of shape (H, W), got {masks.shape}.")

        if masks.dtype != np.int8:
            masks = masks.astype(np.int8)

    detections = []
    for detection_idx in range(n_bboxes):
        detection = ImgDetectionExtended()
        x_center, y_center, width, height = bboxes[detection_idx]
        detection.x_center = x_center.item()
        detection.y_center = y_center.item()
        detection.width = width.item()
        detection.height = height.item()
        detection.confidence = scores[detection_idx].item()

        if angles is not None:
            detection.angle = angles[detection_idx].item()
        if labels is not None:
            detection.label = labels[detection_idx].item()
        if keypoints is not None:
            if keypoints_scores is not None:
                detection.keypoints = transform_to_keypoints(
                    keypoints[detection_idx], keypoints_scores[detection_idx]
                )
            else:
                detection.keypoints = transform_to_keypoints(keypoints[detection_idx])

        detections.append(detection)

    detections_msg = ImgDetectionsExtended()
    detections_msg.detections = detections

    if masks is not None:
        detections_msg.masks = masks

    return detections_msg


def create_line_detection_message(lines: np.ndarray, scores: np.ndarray):
    """Create a DepthAI message for a line detection.

    @param lines: Detected lines of shape (N,4) meaning [...,[x_start, y_start, x_end, y_end],...].
    @type lines: np.ndarray
    @param scores: Confidence scores of detected lines of shape (N,).
    @type scores: np.ndarray

    @return: Message containing the lines and confidence scores of detected lines.
    @rtype: Lines

    @raise ValueError: If the lines are not a numpy array.
    @raise ValueError: If the lines are not of shape (N,4).
    @raise ValueError: If the lines 2nd dimension is not of size E{4}.
    @raise ValueError: If the scores are not a numpy array.
    @raise ValueError: If the scores are not of shape (N,).
    @raise ValueError: If the scores do not have the same length as lines.
    """

    # checks for lines
    if not isinstance(lines, np.ndarray):
        raise ValueError(f"Lines should be numpy array, got {type(lines)}.")
    if len(lines) != 0:
        if len(lines.shape) != 2:
            raise ValueError(
                f"Lines should be of shape (N,4) meaning [...,[x_start, y_start, x_end, y_end],...], got {lines.shape}."
            )
        if lines.shape[1] != 4:
            raise ValueError(
                f"Lines 2nd dimension should be of size 4 e.g. [x_start, y_start, x_end, y_end] got {lines.shape[1]}."
            )

    # checks for scores
    if not isinstance(scores, np.ndarray):
        raise ValueError(f"Scores should be numpy array, got {type(scores)}.")

    if len(scores) != 0:
        if len(scores.shape) != 1:
            raise ValueError(
                f"Scores should be of shape (N,) meaning, got {scores.shape}."
            )

        for score in scores:
            if not isinstance(score, (float, np.floating)):
                raise ValueError(f"Scores should be of type float, got {type(score)}.")

    if scores.shape[0] != lines.shape[0]:
        raise ValueError(
            f"Scores should have same length as lines, got {scores.shape[0]} and {lines.shape[0]}."
        )

    line_detections = []
    for i, line in enumerate(lines):
        line_detection = Line()
        line_detection.start_point = dai.Point2f(line[0], line[1])
        line_detection.end_point = dai.Point2f(line[2], line[3])
        line_detection.confidence = float(scores[i])
        line_detections.append(line_detection)

    lines_msg = Lines()
    lines_msg.lines = line_detections
    return lines_msg



import pytest

from depthai_nodes.ml.messages.creators.clusters import create_cluster_message


def test_non_list_input():
    with pytest.raises(TypeError):
        create_cluster_message(1)


def test_non_list_of_clusters_input():
    with pytest.raises(TypeError):
        create_cluster_message([1, 2, 3])


def test_clusters_input_int():
    clusters = [[(1, 2), (3, 4), (5, 6)], [(7, 8), (9, 10), (11, 12)]]
    create_cluster_message(clusters)


def test_clusters_input_float():
    clusters = [
        [(1.0, 2.0), (3.0, 4.0), (5.0, 6.0)],
        [(7.0, 8.0), (9.0, 10.0), (11.0, 12.0)],
    ]
    create_cluster_message(clusters)


def test_clusters_input_mixed():
    clusters = [[(1, 2), (3, 4), (5, 6)], [(7.0, 8.0), (9.0, 10.0), (11.0, 12.0)]]
    create_cluster_message(clusters)


def test_clusters_input_empty():
    clusters = []
    create_cluster_message(clusters)


def test_clusters_input_empty_cluster():
    clusters = [[]]
    create_cluster_message(clusters)


def test_clusters_input_empty_point():
    with pytest.raises(TypeError):
        clusters = [[[]]]
        create_cluster_message(clusters)


def test_clusters_input_empty_point_value():
    with pytest.raises(TypeError):
        clusters = [[[1, 2], []]]
        create_cluster_message(clusters)


def test_clusters_input_point_shape():
    with pytest.raises(ValueError):
        clusters = [
            [(1, 2, 3), (3, 4, 5), (5, 6, 7)],
            [(7, 8, 9), (9, 10, 11), (11, 12, 13)],
        ]
        create_cluster_message(clusters)



import re

import numpy as np
import pytest

from depthai_nodes.ml.messages import (
    ImgDetectionsExtended,
)
from depthai_nodes.ml.messages.creators.detection import create_detection_message


def test_bboxes_not_numpy_array():
    with pytest.raises(
        ValueError, match="Bounding boxes should be a numpy array, got <class 'int'>."
    ):
        create_detection_message(bboxes=1, scores=1)


def test_bbox_dim():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Bounding boxes should be a 2D array like [... , [x_center, y_center, width, height], ... ], got (1,)."
        ),
    ):
        create_detection_message(bboxes=np.array([1.0]), scores=1)


def test_bbox_dim_4():
    with pytest.raises(
        ValueError,
        match=re.escape("Bounding boxes should be of shape (n, 4), got (1, 3)."),
    ):
        create_detection_message(bboxes=np.array([[0.25, 0.25, 0.25]]), scores=1)


def test_scores_not_numpy_array():
    with pytest.raises(
        ValueError, match="Scores should be a numpy array, got <class 'int'>."
    ):
        create_detection_message(bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]), scores=1)


def test_scores_shape():
    with pytest.raises(
        ValueError,
        match=re.escape("Scores should be of shape (N,), got (1, 1)."),
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]), scores=np.array([[0.1]])
        )


def test_scores_length():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Scores should have same length as bboxes, got 2 scores and 1 bounding boxes."
        ),
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]), scores=np.array([0.1, 0.2])
        )


def test_labels_list():
    with pytest.raises(
        ValueError, match="Labels should be a numpy array, got <class 'int'>."
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            labels=1,
        )


def test_labels_bbox_lengths():
    with pytest.raises(
        ValueError,
        match="Labels should have same length as bboxes, got 2 labels and 1 bounding boxes.",
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            labels=np.array([0.1, 0.2]),
        )


def test_angles_list():
    with pytest.raises(
        ValueError, match="Angles should be a numpy array, got <class 'int'>."
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            angles=1,
        )


def test_angles_bbox_lengths():
    with pytest.raises(
        ValueError,
        match="Angles should have same length as bboxes, got 2 angles and 1 bounding boxes.",
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            angles=np.array([0.1, 0.2]),
        )


def test_angles_values():
    with pytest.raises(
        ValueError, match="Angles should be between -360 and 360, got 999."
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            angles=np.array([999]),
        )


def test_keypoints_list():
    with pytest.raises(
        ValueError, match="Keypoints should be a numpy array, got <class 'int'>."
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            keypoints=1,
        )


def test_keypoints_shape():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Keypoints should be of shape (N, M, 2 or 3) meaning [..., [x, y] or [x, y, z], ...], got (1, 3)."
        ),
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            keypoints=np.array([[1.0, 2.0, 3.0]]),
        )


def test_keypoints_bbox_lengths():
    with pytest.raises(
        ValueError,
        match="Keypoints should have same length as bboxes, got 2 keypoints and 1 bounding boxes",
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            keypoints=np.array([[[1.0, 2.0, 3.0]], [[1.0, 2.0, 3.0]]]),
        )


def test_keypoints_scores_without_keypoints():
    with pytest.raises(
        ValueError,
        match="Keypoints scores should be provided only if keypoints are provided.",
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            keypoints_scores=1,
        )


def test_keypoints_scores_list():
    with pytest.raises(
        ValueError, match="Keypoints scores should be a numpy array, got <class 'int'>."
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            keypoints=np.array([[[1.0, 2.0, 3.0]]]),
            keypoints_scores=1,
        )


def test_keypoints_scores_lengths():
    with pytest.raises(
        ValueError,
        match="Keypoints scores should have same length as keypoints, got 2 keypoints scores and 1 keypoints.",
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            keypoints=np.array([[[1.0, 2.0, 3.0]]]),
            keypoints_scores=np.array([[0.1], [0.2]]),
        )


def test_number_of_keypoints_scores_per_detection():
    with pytest.raises(
        ValueError,
        match="Number of keypoints scores per detection should be the same as number of keypoints per detection, got 2 keypoints scores and 1 keypoints.",
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            keypoints=np.array([[[1.0, 2.0, 3.0]]]),
            keypoints_scores=np.array([[0.1, 0.2]]),
        )


def test_keypoints_scores_values():
    with pytest.raises(ValueError, match="Keypoints scores should be between 0 and 1."):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            keypoints=np.array([[[1.0, 2.0, 3.0]]]),
            keypoints_scores=np.array([[-0.1]]),
        )


def test_masks_list():
    with pytest.raises(
        ValueError, match="Masks should be a numpy array, got <class 'int'>."
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]), scores=np.array([0.1]), masks=1
        )


def test_masks_shape():
    with pytest.raises(
        ValueError,
        match=re.escape("Masks should be of shape (H, W), got (3,)."),
    ):
        create_detection_message(
            bboxes=np.array([[0.25, 0.25, 0.25, 0.25]]),
            scores=np.array([0.1]),
            masks=np.array([1, 2, 3]),
        )


def test_only_bboxes_scores():
    bboxes = np.array(
        [[0.2, 0.2, 0.4, 0.4], [0.5, 0.5, 0.1, 0.1], [0.1, 0.1, 0.2, 0.2]]
    )
    scores = np.array([0.1, 0.2, 0.3])

    message = create_detection_message(bboxes=bboxes, scores=scores)

    assert isinstance(message, ImgDetectionsExtended)
    assert all(detection.label == -1 for detection in message.detections)
    for i, detection in enumerate(message.detections):
        assert detection.x_center == bboxes[i, 0]
        assert detection.y_center == bboxes[i, 1]
        assert detection.width == bboxes[i, 2]
        assert detection.height == bboxes[i, 3]
        assert np.isclose(detection.confidence, scores[i])


def test_bboxes_scores_labels():
    bboxes = np.array(
        [[0.2, 0.2, 0.4, 0.4], [0.5, 0.5, 0.1, 0.1], [0.1, 0.1, 0.2, 0.2]]
    )
    scores = np.array([0.1, 0.2, 0.3])
    labels = np.array([1, 2, 3])

    message = create_detection_message(bboxes=bboxes, scores=scores, labels=labels)

    for i, label in enumerate(labels):
        assert message.detections[i].label == label


def test_bboxes_scores_keypoints():
    bboxes = np.array(
        [[0.2, 0.2, 0.4, 0.4], [0.5, 0.5, 0.1, 0.1], [0.1, 0.1, 0.2, 0.2]]
    )
    scores = np.array([0.1, 0.2, 0.3])
    keypoints = np.array(
        [[(0.0, 0.0), (0.1, 0.1)], [(0.2, 0.2), (0.3, 0.3)], [(0.4, 0.4), (0.5, 0.5)]]
    )

    message = create_detection_message(
        bboxes=bboxes, scores=scores, keypoints=keypoints
    )

    assert isinstance(message, ImgDetectionsExtended)

    for i, detection in enumerate(message.detections):
        for ii, keypoint in enumerate(detection.keypoints):
            assert keypoint.x == keypoints[i][ii][0]
            assert keypoint.y == keypoints[i][ii][1]


if __name__ == "__main__":
    pytest.main()



from typing import List

import depthai as dai


class Keypoint(dai.Buffer):
    """Keypoint class for storing a keypoint.

    Attributes
    ----------
    x: float
        X coordinate of the keypoint, relative to the input height.
    y: float
        Y coordinate of the keypoint, relative to the input width.
    z: Optional[float]
        Z coordinate of the keypoint.
    confidence: Optional[float]
        Confidence of the keypoint.
    """

    def __init__(self):
        """Initializes the Keypoint object."""
        super().__init__()
        self._x: float = None
        self._y: float = None
        self._z: float = 0.0
        self._confidence: float = -1.0

    @property
    def x(self) -> float:
        """Returns the X coordinate of the keypoint.

        @return: X coordinate of the keypoint.
        @rtype: float
        """
        return self._x

    @x.setter
    def x(self, value: float):
        """Sets the X coordinate of the keypoint.

        @param value: X coordinate of the keypoint.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("x must be a float.")
        if value < 0 or value > 1:
            raise ValueError("x must be between 0 and 1.")
        self._x = value

    @property
    def y(self) -> float:
        """Returns the Y coordinate of the keypoint.

        @return: Y coordinate of the keypoint.
        @rtype: float
        """
        return self._y

    @y.setter
    def y(self, value: float):
        """Sets the Y coordinate of the keypoint.

        @param value: Y coordinate of the keypoint.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("y must be a float.")
        if value < 0 or value > 1:
            raise ValueError("y must be between 0 and 1.")
        self._y = value

    @property
    def z(self) -> float:
        """Returns the Z coordinate of the keypoint.

        @return: Z coordinate of the keypoint.
        @rtype: float
        """
        return self._z

    @z.setter
    def z(self, value: float):
        """Sets the Z coordinate of the keypoint.

        @param value: Z coordinate of the keypoint.
        @type value: float
        @raise TypeError: If value is not a float.
        """
        if not isinstance(value, float):
            raise TypeError("z must be a float.")
        self._z = value

    @property
    def confidence(self) -> float:
        """Returns the confidence of the keypoint.

        @return: Confidence of the keypoint.
        @rtype: float
        """
        return self._confidence

    @confidence.setter
    def confidence(self, value: float):
        """Sets the confidence of the keypoint.

        @param value: Confidence of the keypoint.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("confidence must be a float.")
        if value < 0 or value > 1:
            raise ValueError("confidence must be between 0 and 1.")
        self._confidence = value


class Keypoints(dai.Buffer):
    """Keypoints class for storing keypoints.

    Attributes
    ----------
    keypoints: List[Keypoint]
        List of Keypoint objects, each representing a keypoint.
    """

    def __init__(self):
        """Initializes the Keypoints object."""
        super().__init__()
        self._keypoints: List[Keypoint] = []

    @property
    def keypoints(self) -> List[Keypoint]:
        """Returns the keypoints.

        @return: List of keypoints.
        @rtype: List[Keypoint]
        """
        return self._keypoints

    @keypoints.setter
    def keypoints(self, value: List[Keypoint]):
        """Sets the keypoints.

        @param value: List of keypoints.
        @type value: List[Keypoint]
        @raise TypeError: If value is not a list.
        @raise TypeError: If each each element is not of type Keypoint.
        """
        if not isinstance(value, list):
            raise TypeError("keypoints must be a list.")
        if not all(isinstance(item, Keypoint) for item in value):
            raise ValueError("keypoints must be a list of Keypoint objects.")
        self._keypoints = value



from typing import Any, Dict, List, Tuple

import depthai as dai
import numpy as np

from ..messages.creators import create_cluster_message
from .base_parser import BaseParser
from .utils.ufld import decode_ufld


class LaneDetectionParser(BaseParser):
    """
    Parser class for Ultra-Fast-Lane-Detection model. It expects one ouput layer containing the lane detection results.
    It supports two versions of the model: CuLane and TuSimple. Results are representented with clusters of points.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    row_anchors : List[int]
        List of row anchors.
    griding_num : int
        Griding number.
    cls_num_per_lane : int
        Number of points per lane.
    input_shape : Tuple[int, int]
        Input shape.

    Output Message/s
    ----------------
    **Type**: Clusters
    **Description**: Detected lanes represented as clusters of points.

    Error Handling
    --------------
    **ValueError**: If the row anchors are not specified.
    **ValueError**: If the griding number is not specified.
    **ValueError**: If the number of points per lane is not specified.
    """

    def __init__(
        self,
        output_layer_name: str = "",
        row_anchors: List[int] = None,
        griding_num: int = None,
        cls_num_per_lane: int = None,
        input_shape: Tuple[int, int] = (288, 800),
    ) -> None:
        """Initializes the lane detection parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param row_anchors: List of row anchors.
        @type row_anchors: List[int]
        @param griding_num: Griding number.
        @type griding_num: int
        @param cls_num_per_lane: Number of points per lane.
        @type cls_num_per_lane: int
        @param input_shape: Input shape.
        @type input_shape: Tuple[int, int]
        """
        super().__init__()
        self.output_layer_name = output_layer_name

        self.row_anchors = row_anchors
        self.griding_num = griding_num
        self.cls_num_per_lane = cls_num_per_lane
        self.input_shape = input_shape

    def setOutputLayerName(self, output_layer_name: str) -> None:
        """Set the output layer name for the lane detection model.

        @param output_layer_name: Name of the output layer.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name

    def setRowAnchors(self, row_anchors: List[int]) -> None:
        """Set the row anchors for the lane detection model.

        @param row_anchors: List of row anchors.
        @type row_anchors: List[int]
        """
        if not isinstance(row_anchors, list):
            raise ValueError("Row anchors must be a list.")
        if not all(isinstance(anchor, int) for anchor in row_anchors):
            raise ValueError("Row anchors must be a list of integers.")
        self.row_anchors = row_anchors

    def setGridingNum(self, griding_num: int) -> None:
        """Set the griding number for the lane detection model.

        @param griding_num: Griding number.
        @type griding_num: int
        """
        if not isinstance(griding_num, int):
            raise ValueError("Griding number must be an integer.")
        self.griding_num = griding_num

    def setClsNumPerLane(self, cls_num_per_lane: int) -> None:
        """Set the number of points per lane for the lane detection model.

        @param cls_num_per_lane: Number of classes per lane.
        @type cls_num_per_lane: int
        """
        if not isinstance(cls_num_per_lane, int):
            raise ValueError("Number of points per lane must be an integer.")
        self.cls_num_per_lane = cls_num_per_lane

    def setInputShape(self, input_shape: Tuple[int, int]) -> None:
        """Set the input shape for the lane detection model.

        @param input_shape: Input shape.
        @type input_shape: Tuple[int, int]
        """
        if not isinstance(input_shape, tuple):
            raise ValueError("Input shape must be a tuple.")
        if len(input_shape) != 2:
            raise ValueError("Input shape must be a tuple of two integers.")
        if not all(isinstance(size, int) for size in input_shape):
            raise ValueError("Input shape must be a tuple of integers.")
        self.input_shape = input_shape

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "LaneDetectionParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: LaneDetectionParser
        """

        output_layers = head_config.get("outputs", [])
        if len(output_layers) != 1:
            raise ValueError(
                f"Only one output layer supported for LaneDetectionParser, got {len(output_layers)} layers."
            )
        self.output_layer_name = output_layers[0]
        self.row_anchors = head_config.get("row_anchors", self.row_anchors)
        self.griding_num = head_config.get("griding_num", self.griding_num)
        self.cls_num_per_lane = head_config.get(
            "cls_num_per_lane", self.cls_num_per_lane
        )

        return self

    def run(self):
        if self.row_anchors is None:
            raise ValueError("Row anchors must be specified!")
        if self.griding_num is None:
            raise ValueError("Griding number must be specified!")
        if self.cls_num_per_lane is None:
            raise ValueError("Number of points per lane must be specified!")

        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            tensor = output.getTensor(self.output_layer_name, dequantize=True).astype(
                np.float32
            )
            y = tensor[0]

            points = decode_ufld(
                anchors=self.row_anchors,
                griding_num=self.griding_num,
                cls_num_per_lane=self.cls_num_per_lane,
                input_width=self.input_shape[1],
                input_height=self.input_shape[0],
                y=y,
            )

            message = create_cluster_message(points)
            message.setTimestamp(output.getTimestamp())
            self.out.send(message)



from typing import List

import depthai as dai


class Cluster(dai.Buffer):
    """Cluster class for storing a cluster.

    Attributes
    ----------
    label : int
        Label of the cluster.
    points : List[dai.Point2f]
        List of points in the cluster.
    """

    def __init__(self):
        """Initializes the Cluster object."""
        super().__init__()
        self._label: int = None
        self.points: List[dai.Point2f] = []

    @property
    def label(self) -> int:
        """Returns the label of the cluster.

        @return: Label of the cluster.
        @rtype: int
        """
        return self._label

    @label.setter
    def label(self, value: int):
        """Sets the label of the cluster.

        @param value: Label of the cluster.
        @type value: int
        @raise TypeError: If value is not an int.
        """
        if not isinstance(value, int):
            raise TypeError(f"Label must be of type int, instead got {type(value)}.")
        self._label = value

    @property
    def points(self) -> List[dai.Point2f]:
        """Returns the points in the cluster.

        @return: List of points in the cluster.
        @rtype: List[dai.Point2f]
        """
        return self._points

    @points.setter
    def points(self, value: List[dai.Point2f]):
        """Sets the points in the cluster.

        @param value: List of points in the cluster.
        @type value: List[dai.Point2f]
        @raise TypeError: If value is not a list.
        @raise TypeError: If each element is not of type dai.Point2f.
        """
        if not isinstance(value, List):
            raise TypeError(f"Points must be a list, instead got {type(value)}.")
        if not all(isinstance(point, dai.Point2f) for point in value):
            raise ValueError("Points must be a list of dai.Point2f objects")
        self._points = value


class Clusters(dai.Buffer):
    """Clusters class for storing clusters.

    Attributes
    ----------
    clusters : List[Cluster]
        List of clusters.
    """

    def __init__(self):
        """Initializes the Clusters object."""
        super().__init__()
        self._clusters: List[Cluster] = []

    @property
    def clusters(self) -> List[Cluster]:
        """Returns the clusters.

        @return: List of clusters.
        @rtype: List[Cluster]
        """
        return self._clusters

    @clusters.setter
    def clusters(self, value: List[Cluster]):
        """Sets the clusters.

        @param value: List of clusters.
        @type value: List[Cluster]
        @raise TypeError: If value is not a list.
        @raise ValueError: If each element is not of type Cluster.
        """
        if not isinstance(value, List):
            raise TypeError("Clusters must be a list.")
        if not all(isinstance(cluster, Cluster) for cluster in value):
            raise ValueError("Clusters must be a list of Cluster objects.")
        self._clusters = value



from typing import Any, Dict

import depthai as dai
import numpy as np

from ..messages.creators import create_keypoints_message
from .base_parser import BaseParser


class KeypointParser(BaseParser):
    """Parser class for 2D or 3D keypoints models. It expects one ouput layer containing
    keypoints. The number of keypoints must be specified. Moreover, the keypoints are
    normalized by a scale factor if provided.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    scale_factor : float
        Scale factor to divide the keypoints by.
    n_keypoints : int
        Number of keypoints the model detects.
    score_threshold : float
        Confidence score threshold for detected keypoints.

    Output Message/s
    ----------------
    **Type**: Keypoints

    **Description**: Keypoints message containing 2D or 3D keypoints.

    Error Handling
    --------------
    **ValueError**: If the number of keypoints is not specified.

    **ValueError**: If the number of coordinates per keypoint is not 2 or 3.

    **ValueError**: If the number of output layers is not 1.
    """

    def __init__(
        self,
        output_layer_name: str = "",
        scale_factor: float = 1.0,
        n_keypoints: int = None,
        score_threshold: float = None,
    ) -> None:
        """Initializes the parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param scale_factor: Scale factor to divide the keypoints by.
        @type scale_factor: float
        @param n_keypoints: Number of keypoints.
        @type n_keypoints: int
        """
        super().__init__()
        self.output_layer_name = output_layer_name
        self.scale_factor = scale_factor
        self.n_keypoints = n_keypoints
        self.score_threshold = score_threshold

    def setOutputLayerName(self, output_layer_name: str) -> None:
        """Sets the name of the output layer.

        @param output_layer_name: The name of the output layer.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name

    def setScaleFactor(self, scale_factor: float) -> None:
        """Sets the scale factor to divide the keypoints by.

        @param scale_factor: Scale factor to divide the keypoints by.
        @type scale_factor: float
        """
        if not isinstance(scale_factor, float):
            raise ValueError("Scale factor must be a float.")
        self.scale_factor = scale_factor

    def setNumKeypoints(self, n_keypoints: int) -> None:
        """Sets the number of keypoints.

        @param n_keypoints: Number of keypoints.
        @type n_keypoints: int
        """
        if not isinstance(n_keypoints, int):
            raise ValueError("Number of keypoints must be an integer.")
        self.n_keypoints = n_keypoints

    def setScoreThreshold(self, threshold: float) -> None:
        """Sets the confidence score threshold for the detected body keypoints.

        @param threshold: Confidence score threshold for detected keypoints.
        @type threshold: float
        """
        if not isinstance(threshold, float):
            raise ValueError("Confidence threshold must be a float.")
        self.score_threshold = threshold

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "KeypointParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: KeypointParser
        """

        output_layers = head_config["outputs"]
        if len(output_layers) != 1:
            raise ValueError(
                f"Only one output layer supported for Keypoint, got {output_layers} layers."
            )
        self.output_layer_name = output_layers[0]
        self.scale_factor = head_config.get("scale_factor", self.scale_factor)
        self.n_keypoints = head_config.get("n_keypoints", self.n_keypoints)
        self.score_threshold = head_config.get("score_threshold", self.score_threshold)

        return self

    def run(self):
        if self.n_keypoints is None:
            raise ValueError("Number of keypoints must be specified!")

        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            keypoints = output.getTensor(
                self.output_layer_name, dequantize=True
            ).astype(np.float32)
            num_coords = int(np.prod(keypoints.shape) / self.n_keypoints)

            if num_coords not in [2, 3]:
                raise ValueError(
                    f"Expected 2 or 3 coordinates per keypoint, got {num_coords}."
                )

            keypoints = keypoints.reshape(self.n_keypoints, num_coords)

            keypoints /= self.scale_factor

            keypoints = np.clip(keypoints, 0, 1)

            msg = create_keypoints_message(keypoints)
            msg.setTimestamp(output.getTimestamp())

            self.out.send(msg)



import cv2
import depthai as dai

from depthai_nodes.ml.messages import Classifications

from .utils.message_parsers import (
    parse_classification_message,
)


def visualize_classification(
    frame: dai.ImgFrame, message: Classifications, extraParams: dict
):
    """Visualizes the classification on the frame."""
    classes, scores = parse_classification_message(message)
    classes = classes[:2]
    scores = scores[:2]
    if frame.shape[0] < 128:
        frame = cv2.resize(frame, (frame.shape[1] * 2, frame.shape[0] * 2))
    for i, (cls, score) in enumerate(zip(classes, scores)):
        cv2.putText(
            frame,
            f"{cls}: {score:.2f}",
            (10, 20 + 20 * i),
            cv2.FONT_HERSHEY_TRIPLEX,
            0.5,
            255,
        )

    cv2.imshow("Classification", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False


def visualize_text_recognition(
    frame: dai.ImgFrame, message: Classifications, extraParams: dict
):
    """Visualizes the text recognition on the frame."""

    if frame.shape[0] < 128:
        frame = cv2.resize(frame, (frame.shape[1] * 2, frame.shape[0] * 2))

    classes, scores = parse_classification_message(message)
    text = "".join(classes)
    cv2.putText(frame, text, (10, 20), cv2.FONT_HERSHEY_TRIPLEX, 0.5, 255)

    cv2.imshow("Text recognition", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False



from typing import List

import depthai as dai
import numpy as np
from numpy.typing import NDArray


class Classifications(dai.Buffer):
    """Classification class for storing the classes and their respective scores.

    Attributes
    ----------
    classes : list[str]
        A list of classes.
    scores : NDArray[np.float32]
        Corresponding probability scores.
    """

    def __init__(self):
        """Initializes the Classifications object."""
        dai.Buffer.__init__(self)
        self._classes: List[str] = []
        self._scores: NDArray[np.float32] = np.array([])

    @property
    def classes(self) -> List:
        """Returns the list of classes.

        @return: List of classes.
        @rtype: List[str]
        """
        return self._classes

    @classes.setter
    def classes(self, value: List[str]):
        """Sets the classes.

        @param value: A list of class names.
        @type value: List[str]
        @raise TypeError: If value is not a list.
        @raise ValueError: If each element is not of type string.
        """
        if not isinstance(value, List):
            raise TypeError(f"Classes must be a list, instead got {type(value)}.")
        if not all(isinstance(class_name, str) for class_name in value):
            raise ValueError("Classes must be a list of strings.")
        self._classes = value

    @property
    def scores(self) -> NDArray:
        """Returns the list of scores.

        @return: List of scores.
        @rtype: NDArray[np.float32]
        """
        return self._scores

    @scores.setter
    def scores(self, value: NDArray[np.float32]):
        """Sets the scores.

        @param value: A list of scores.
        @type value: NDArray[np.float32]
        @raise TypeError: If value is not a numpy array.
        @raise ValueError: If value is not a 1D numpy array.
        @raise ValueError: If each element is not of type float.
        """
        if not isinstance(value, np.ndarray):
            raise TypeError(f"Scores must be a np.ndarray, instead got {type(value)}.")
        if value.ndim != 1:
            raise ValueError("Scores must be a 1D a np.ndarray.")
        if value.size > 0 and value.dtype != np.float32:
            raise ValueError("Scores must be a np.ndarray of floats.")
        self._scores = value



from typing import Any, Dict, List

import depthai as dai

from .base_parser import BaseParser


class EmbeddingsParser(BaseParser):
    """Parser class for parsing the output of embeddings neural network model head.

    Attributes
    ----------

    Output Message/s
    ----------------
    **Type**: ImgDetectionsExtended

    **Description**: Message containing bounding boxes, labels, confidence scores, and keypoints or masks and protos of the detected objects.
    """

    def __init__(self) -> None:
        """Initialize the EmbeddingsParser node."""
        super().__init__()
        self.output_layer_names = None

    def build(self, head_config: Dict[str, Any]):
        """Sets the head configuration for the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: EmbeddingsParser
        """

        self.output_layer_names = head_config["outputs"]
        assert (
            len(self.output_layer_names) == 1
        ), "Embeddings head should have only one output layer"

        return self

    def setOutputLayerNames(self, output_layer_names: List[str]) -> None:
        """Sets the output layer names for the parser.

        @param output_layer_names: The output layer names for the parser.
        @type output_layer_names: List[str]
        """
        self.output_layer_names = output_layer_names

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped, no more data
            # Get all the layer names
            output_names = self.output_layer_names or output.getAllLayerNames()
            assert (
                len(output_names) == 1
            ), "Embeddings head should have only one output layer"

            self.out.send(output)



import numpy as np
import pytest

from depthai_nodes.ml.messages import Predictions
from depthai_nodes.ml.messages.creators.regression import create_regression_message

np.random.seed(0)


def test_not_list_predictions():
    predictions = 10
    with pytest.raises(
        ValueError, match="Predictions should be list, got <class 'int'>."
    ):
        create_regression_message(predictions)


def test_not_float_predictions():
    predictions = np.random.randint(0, 10, 5).tolist()
    with pytest.raises(
        ValueError,
        match="Each prediction should be a float, got <class 'int'> instead.",
    ):
        create_regression_message(predictions)


def test_predictions_object():
    predictions = np.random.rand(10).tolist()

    message = create_regression_message(predictions)

    assert isinstance(message, Predictions)
    assert predictions[0] == message.predictions[0].prediction


if __name__ == "__main__":
    pytest.main()



import re

import numpy as np
import pytest

from depthai_nodes.ml.messages import Line, Lines
from depthai_nodes.ml.messages.creators.detection import create_line_detection_message


def test_not_numpy_lines():
    with pytest.raises(
        ValueError, match="Lines should be numpy array, got <class 'list'>."
    ):
        create_line_detection_message([1, 2, 3], [0.1, 0.2, 0.3])


def test_empty_numpy_lines():
    lines_msg = create_line_detection_message(np.array([]), np.array([]))
    assert isinstance(lines_msg, Lines)
    assert len(lines_msg.lines) == 0


def test_lines_shape():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Lines should be of shape (N,4) meaning [...,[x_start, y_start, x_end, y_end],...], got (1,)."
        ),
    ):
        create_line_detection_message(np.array([1]), [0.1])


def test_lines_dim_4():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Lines 2nd dimension should be of size 4 e.g. [x_start, y_start, x_end, y_end] got 3."
        ),
    ):
        create_line_detection_message(np.array([[1, 2, 3]]), [0.1])


def test_not_numpy_scores():
    with pytest.raises(
        ValueError, match="Scores should be numpy array, got <class 'list'>."
    ):
        create_line_detection_message(np.array([[1, 2, 3, 4]]), [0.1, 0.2, 0.3])


def test_empty_numpy_scores():
    with pytest.raises(
        ValueError, match="Scores should have same length as lines, got 0 and 1."
    ):
        create_line_detection_message(np.array([[1, 2, 3, 4]]), np.array([]))


def test_mixed_scores():
    with pytest.raises(
        ValueError,
        match="Scores should be of type float, got <class 'NoneType'>.",
    ):
        create_line_detection_message(
            np.array([[1, 2, 3, 4], [1, 2, 3, 4]]), np.array([0.1, None])
        )


def test_scores_shape():
    with pytest.raises(
        ValueError,
        match=re.escape("Scores should be of shape (N,) meaning, got (1, 1)."),
    ):
        create_line_detection_message(np.array([[1, 2, 3, 4]]), np.array([[0.1]]))


def test_scores_length():
    with pytest.raises(
        ValueError,
        match=re.escape("Scores should have same length as lines, got 1 and 2."),
    ):
        create_line_detection_message(
            np.array([[1, 2, 3, 4], [1, 2, 3, 4]]), np.array([0.1])
        )


def test_line_detection():
    lines = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
    scores = np.array([0.4, 0.2])
    lines_msg = create_line_detection_message(lines, scores)

    assert isinstance(lines_msg, Lines)
    assert len(lines_msg.lines) == 2

    for i, line in enumerate(lines_msg.lines):
        assert isinstance(line, Line)
        assert line.start_point.x == lines[i][0]
        assert line.start_point.y == lines[i][1]
        assert line.end_point.x == lines[i][2]
        assert line.end_point.y == lines[i][3]
        assert line.confidence == scores[i]


if __name__ == "__main__":
    pytest.main()



import cv2
import depthai as dai
import numpy as np

from depthai_nodes.ml.messages import Map2D

from .utils.message_parsers import parse_map_message


def visualize_map(frame: dai.ImgFrame, message: Map2D, extraParams: dict):
    """Visualizes the map on the frame."""

    map = parse_map_message(message)

    # make color representation of the map
    map_normalized = cv2.normalize(map, None, 0, 255, cv2.NORM_MINMAX)
    map_normalized = map_normalized.astype(np.uint8)
    colored_map = cv2.applyColorMap(map_normalized, cv2.COLORMAP_INFERNO)
    frame_height, frame_width, _ = frame.shape
    colored_map = cv2.resize(
        colored_map, (frame_width, frame_height), interpolation=cv2.INTER_LINEAR
    )

    alpha = 0.6
    overlay = cv2.addWeighted(colored_map, alpha, frame, 1 - alpha, 0)

    cv2.imshow("Map Overlay", overlay)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False



import re

import depthai as dai
import numpy as np
import pytest

from depthai_nodes.ml.messages.creators.tracked_features import (
    create_tracked_features_message,
)


def test_reference_point_shape():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "reference_points should be of shape (N,2) meaning [...,[x, y],...], got (3,)."
        ),
    ):
        create_tracked_features_message(
            np.array([0.7, 0.3, 0.5]), target_points=np.array([[1, 2], [3, 4]])
        )


def test_reference_point_size():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "reference_points 2nd dimension should be of size 2 e.g. [x, y], got 1."
        ),
    ):
        create_tracked_features_message(
            np.array([[0.7], [0.8]]), target_points=np.array([[1, 2], [3, 4]])
        )


def test_target_point_shape():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "target_points should be of shape (N,2) meaning [...,[x, y],...], got (3,)."
        ),
    ):
        create_tracked_features_message(
            np.array([[0.7, 0.3], [0.5, 0.6]]), target_points=np.array([0.7, 0.3, 0.5])
        )


def test_target_point_size():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "target_points 2nd dimension should be of size 2 e.g. [x, y], got 1."
        ),
    ):
        create_tracked_features_message(
            np.array([[0.7, 0.3], [0.5, 0.6]]), target_points=np.array([[1], [2]])
        )


def test_same_length():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "The number of reference points and target points should be the same."
        ),
    ):
        create_tracked_features_message(
            np.array([[0.7, 0.3], [0.5, 0.6], [0.5, 0.5], [0.6, 0.5]]),
            target_points=np.array([[1, 2], [3, 4]]),
        )


def test_return():
    reference_points = np.array([[0.7, 0.3], [0.5, 0.6], [0.5, 0.5], [0.6, 0.5]])
    target_points = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])

    msg = create_tracked_features_message(reference_points, target_points)

    assert isinstance(msg, dai.TrackedFeatures)
    assert len(msg.trackedFeatures) == 8
    for i in range(4):
        assert isinstance(msg.trackedFeatures[2 * i], dai.TrackedFeature)
        assert np.isclose(msg.trackedFeatures[2 * i].position.x, reference_points[i][0])
        assert np.isclose(msg.trackedFeatures[2 * i].position.y, reference_points[i][1])
        assert msg.trackedFeatures[2 * i].id == i
        assert msg.trackedFeatures[2 * i].age == 0
        assert isinstance(msg.trackedFeatures[2 * i + 1], dai.TrackedFeature)
        assert np.isclose(
            msg.trackedFeatures[2 * i + 1].position.x, target_points[i][0]
        )
        assert np.isclose(
            msg.trackedFeatures[2 * i + 1].position.y, target_points[i][1]
        )
        assert msg.trackedFeatures[2 * i + 1].id == i
        assert msg.trackedFeatures[2 * i + 1].age == 1


if __name__ == "__main__":
    pytest.main()



from typing import List

import numpy as np

from ...messages.creators.keypoints import Keypoint


def normalize_keypoints(keypoints: np.ndarray, height: int, width: int) -> np.ndarray:
    """Normalize keypoint coordinates to (0, 1).

    Parameters:
    @param keypoints: A numpy array of shape (N, 2) or (N, K, 2) where N is the number of keypoint sets and K is the number of keypoint in each set.
    @type np.ndarray
    @param height: The height of the image.
    @type height: int
    @param width: The width of the image.
    @type width: int

    Returns:
    np.ndarray: A numpy array of shape (N, 2) containing the normalized keypoints.
    """
    keypoints = keypoints.astype(np.float32)
    if not isinstance(keypoints, np.ndarray):
        raise TypeError("Keypoints must be a numpy array.")

    if len(keypoints.shape) != 2 and len(keypoints.shape) != 3:
        raise ValueError(
            f"Keypoints must be of shape (N, 2) or (N, K, 2). Got {keypoints.shape}."
        )

    if keypoints.shape[1] != 2 and len(keypoints.shape) == 2:
        raise ValueError(
            "Keypoints must be of shape (N, 2). Other options are currently not supported"
        )
    elif len(keypoints.shape) == 3:
        if keypoints.shape[2] != 2:
            raise ValueError(
                "Keypoints must be of shape (N, K, 2). Other options are currently not supported"
            )
    keypoints[:, 0] = keypoints[:, 0] / width
    keypoints[:, 1] = keypoints[:, 1] / height

    return keypoints


def transform_to_keypoints(
    keypoints: np.ndarray, confidences: np.ndarray = None
) -> List[Keypoint]:
    """Transforms an array representing keypoints into a list of Keypoint objects.

    @param keypoints: Detected 2D or 3D keypoints of shape (N,2 or 3) meaning [...,[x, y],...] or [...,[x, y, z],...].
    @type keypoints: np.ndarray
    @param confidences: Confidence scores of the detected keypoints.
    @type confidences: Optional[np.ndarray]

    @return: List of Keypoint objects.
    @rtype: List[Keypoint]
    """
    if not isinstance(keypoints, np.ndarray):
        raise ValueError("Keypoints must be a numpy array.")
    if len(keypoints.shape) != 2:
        raise ValueError(
            "Keypoints must be of shape (N, 2). Got shape {keypoints.shape}."
        )
    if keypoints.shape[1] != 2 and keypoints.shape[1] != 3:
        raise ValueError(
            "Keypoints must be of shape (N, 2) or (N, 3). Got shape {keypoints.shape}."
        )

    if confidences is not None:
        if not isinstance(confidences, np.ndarray):
            raise ValueError("Confidences must be a numpy array.")
        if len(confidences.shape) != 1:
            raise ValueError(
                "Confidences must be of shape (N,). Got shape {confidences.shape}."
            )
        if len(confidences) != len(keypoints):
            raise ValueError(
                "Confidences should have same length as keypoints, got {len(confidences)} confidences and {len(keypoints)} keypoints."
            )

    dim = keypoints.shape[1]

    keypoints_list = []
    for i, keypoint in enumerate(keypoints):
        kp = Keypoint()
        kp.x = keypoint[0]
        kp.y = keypoint[1]

        if dim == 3:
            kp.z = keypoint[2]
        if confidences is not None:
            kp.confidence = confidences[i]

        keypoints_list.append(kp)

    return keypoints_list



from typing import List

import depthai as dai


class Prediction(dai.Buffer):
    """Prediction class for storing a prediction.

    Attributes
    ----------
    prediction : float
        The predicted value.
    """

    def __init__(self):
        """Initializes the Prediction object."""
        super().__init__()
        self._prediction: float = None

    @property
    def prediction(self) -> float:
        """Returns the prediction.

        @return: The predicted value.
        @rtype: float
        """
        return self._prediction

    @prediction.setter
    def prediction(self, value: float):
        """Sets the prediction.

        @param value: The predicted value.
        @type value: float
        @raise TypeError: If value is not of type float.
        """
        if not isinstance(value, float):
            raise TypeError(
                f"Prediction must be of type float, instead got {type(value)}."
            )
        self._prediction = value


class Predictions(dai.Buffer):
    """Predictions class for storing predictions.

    Attributes
    ----------
    predictions : List[Prediction]
        List of predictions.
    """

    def __init__(self):
        """Initializes the Predictions object."""
        super().__init__()
        self._predictions: List[Prediction] = []

    @property
    def predictions(self) -> List[Prediction]:
        """Returns the predictions.

        @return: List of predictions.
        @rtype: List[Prediction]
        """
        return self._predictions

    @predictions.setter
    def predictions(self, value: List[Prediction]):
        """Sets the predictions.

        @param value: List of predicted values.
        @type value: List[Prediction]
        @raise TypeError: If value is not a list.
        @raise ValueError: If each element is not of type Prediction.
        """
        if not isinstance(value, List):
            raise TypeError(
                f"Predictions must be of type list, instead got {type(value)}."
            )
        if not all(isinstance(item, Prediction) for item in value):
            raise ValueError("Predictions must be a list of Prediction objects.")
        self._predictions = value



import argparse
from typing import Tuple

import depthai as dai

from depthai_nodes.parsing_neural_network import ParsingNeuralNetwork


def parse_model_slug(full_slug) -> Tuple[str, str]:
    if ":" not in full_slug:
        raise NameError(
            "Please provide the model slug in the format of 'model_slug:model_version_slug'"
        )
    model_slug_parts = full_slug.split(":")
    model_slug = model_slug_parts[0]
    model_version_slug = model_slug_parts[1]

    return model_slug, model_version_slug


parser = argparse.ArgumentParser()
parser.add_argument(
    "-nn", "--nn_archive", type=str, default=None, help="Path to the NNArchive."
)
parser.add_argument(
    "-s", "--model_slug", type=str, default=None, help="Slug of the model from HubAI."
)
parser.add_argument("-ip", type=str, default="", help="IP of the device")
args = parser.parse_args()

if not (args.nn_archive or args.model_slug):
    raise ValueError("You have to pass either path to NNArchive or model slug")

device = dai.Device(dai.DeviceInfo(args.ip))
with dai.Pipeline(device) as pipeline:
    camera_node = pipeline.create(dai.node.Camera).build()

    if args.model_slug:
        model_slug, model_version_slug = parse_model_slug(args.model_slug)
        model = dai.NNModelDescription(
            modelSlug=model_slug, modelVersionSlug=model_version_slug
        )

    else:
        model = dai.NNArchive(args.nn_archive)

    nn_w_parser = pipeline.create(ParsingNeuralNetwork).build(camera_node, model)

    head_indices = nn_w_parser._parsers.keys()

    parser_output_queues = {
        i: nn_w_parser.getOutput(i).createOutputQueue() for i in head_indices
    }

    pipeline.start()

    while pipeline.isRunning():
        for head_id in parser_output_queues:
            parser_output = parser_output_queues[head_id].get()
            print(f"{head_id} - {type(parser_output)}")
        pipeline.stop()



from pathlib import Path
from typing import Dict, Union, overload

import depthai as dai

from .ml.parsers import BaseParser
from .parser_generator import ParserGenerator


class ParsingNeuralNetwork(dai.node.ThreadedHostNode):
    Propeties = dai.node.NeuralNetwork.Properties
    """Node that wraps the NeuralNetwork node and adds parsing capabilities. A
    NeuralNetwork node is created with it's appropriate parser nodes for each model
    head. Parser nodes are chosen based on the supplied NNArchive.

    Attributes
    ----------
    input : Node.Input
        Neural network input.
    inputs : Node.InputMap
        Neural network inputs.
    out : Node.Output
        Neural network output. Can be used only when there is exactly one model head. Otherwise, getOutput method must be used.
    passthrough : Node.Output
        Neural network passthrough.
    passthroughs : Node.OutputMap
        Neural network passthroughs.
    """

    def __init__(self, *args, **kwargs) -> None:
        """Initializes the ParsingNeuralNetwork node. NeuralNetwork node is created in
        the pipeline.

        @param args: Arguments to be passed to the ThreadedHostNode class.
        @param kwargs: Keyword arguments to be passed to the ThreadedHostNode class.
        """
        super().__init__(*args, **kwargs)
        self._pipeline = self.getParentPipeline()
        self._nn = self._pipeline.create(dai.node.NeuralNetwork)
        self._parsers: Dict[int, BaseParser] = {}

    @overload
    def build(
        self, input: dai.Node.Output, nn_source: dai.NNModelDescription, fps: int
    ) -> "ParsingNeuralNetwork":
        ...

    @overload
    def build(
        self, input: dai.Node.Output, nn_source: dai.NNArchive, fps: int
    ) -> "ParsingNeuralNetwork":
        ...

    def build(
        self,
        input: dai.Node.Output,
        nn_source: Union[dai.NNModelDescription, dai.NNArchive],
        fps: int = None,
    ) -> "ParsingNeuralNetwork":
        """Builds the underlying NeuralNetwork node and creates parser nodes for each
        model head.

        @param input: Node's input. It is a linking point to which the NeuralNetwork is
            linked. It accepts the output of a Camera node.
        @type input: Node.Input
        @param nn_source: NNModelDescription object containing the HubAI model
            descriptors, or NNArchive object of the model.
        @type nn_source: Union[dai.NNModelDescription, dai.NNArchive]
        @param fps: FPS limit for the model runtime.
        @type fps: int
        @return: Returns the ParsingNeuralNetwork object.
        @rtype: ParsingNeuralNetwork
        @raise ValueError: If the nn_source is not a NNModelDescription or NNArchive
            object.
        """

        if isinstance(nn_source, dai.NNModelDescription):
            if not nn_source.platform:
                nn_source.platform = (
                    self.getParentPipeline().getDefaultDevice().getPlatformAsString()
                )
            self._nn_archive = dai.NNArchive(dai.getModelFromZoo(nn_source))
        elif isinstance(nn_source, dai.NNArchive):
            self._nn_archive = nn_source
        else:
            raise ValueError(
                "nn_source must be either a NNModelDescription or NNArchive"
            )

        kwargs = {"fps": fps} if fps else {}
        self._nn.build(input, nn_source, **kwargs)

        self._updateParsers(self._nn_archive)
        return self

    def _updateParsers(self, nnArchive: dai.NNArchive) -> None:
        self._removeOldParserNodes()
        self._parsers = self._getParserNodes(nnArchive)

    def _removeOldParserNodes(self) -> None:
        for parser in self._parsers.values():
            self._pipeline.remove(parser)

    def _getParserNodes(self, nnArchive: dai.NNArchive) -> Dict[int, BaseParser]:
        parser_generator = self._pipeline.create(ParserGenerator)
        parsers = parser_generator.build(nnArchive)
        for parser in parsers.values():
            self._nn.out.link(
                parser.input
            )  # TODO: once NN node has output dictionary, link to the correct output
        self._pipeline.remove(parser_generator)
        return parsers

    def getNumInferenceThreads(self) -> int:
        """Returns number of inference threads of the NeuralNetwork node."""
        return self._nn.getNumInferenceThreads()

    def setBackend(self, setBackend: str) -> None:
        """Sets the backend of the NeuralNetwork node."""
        self._nn.setBackend(setBackend)

    def setBackendProperties(self, setBackendProperties: Dict[str, str]) -> None:
        """Sets the backend properties of the NeuralNetwork node."""
        self._nn.setBackendProperties(setBackendProperties)

    def setBlob(self, blob: Union[Path, dai.OpenVINO.Blob]) -> None:
        """Sets the blob of the NeuralNetwork node."""
        self._nn.setBlob(blob)

    def setBlobPath(self, path: Path) -> None:
        """Sets the blob path of the NeuralNetwork node."""
        self._nn.setBlobPath(path)

    def setFromModelZoo(
        self, description: dai.NNModelDescription, useCached: bool
    ) -> None:
        """Sets the model from the model zoo of the NeuralNetwork node."""
        self._nn.setFromModelZoo(description, useCached)

    def setModelPath(self, modelPath: Path) -> None:
        """Sets the model path of the NeuralNetwork node."""
        self._nn.setModelPath(modelPath)

    def setNNArchive(self, nnArchive: dai.NNArchive) -> None:
        """Sets the NNArchive of the ParsingNeuralNetwork node.

        Updates the NeuralNetwork node and parser nodes.
        """
        self._nn_archive = nnArchive
        self._nn.setNNArchive(nnArchive)
        self._updateParsers(nnArchive)

    def setNumInferenceThreads(self, numThreads: int) -> None:
        """Sets the number of inference threads of the NeuralNetwork node."""
        self._nn.setNumInferenceThreads(numThreads)

    def setNumNCEPerInferenceThread(self, numNCEPerThread: int) -> None:
        """Sets the number of NCE per inference thread of the NeuralNetwork node."""
        self._nn.setNumNCEPerInferenceThread(numNCEPerThread)

    def setNumPoolFrames(self, numFrames: int) -> None:
        """Sets the number of pool frames of the NeuralNetwork node."""
        self._nn.setNumPoolFrames(numFrames)

    def setNumShavesPerInferenceThread(self, numShavesPerInferenceThread: int) -> None:
        """Sets the number of shaves per inference thread of the NeuralNetwork node."""
        self._nn.setNumShavesPerInferenceThread(numShavesPerInferenceThread)

    @property
    def input(self) -> dai.Node.Input:
        return self._nn.input

    @property
    def inputs(self) -> dai.Node.InputMap:
        return self._nn.inputs

    @property
    def out(self) -> dai.Node.Output:
        self._checkNNArchive()
        if len(self._parsers) != 1:
            raise RuntimeError(
                f"Property out is only available when there is exactly one model head. \
                               The model has {len(self._getModelHeads())} heads. Use {self.getOutput.__name__} method instead."
            )
        return list(self._parsers.values())[0].out

    def _getModelHeads(self):
        return self._getConfig().model.heads

    @property
    def passthrough(self) -> dai.Node.Output:
        return self._nn.passthrough

    @property
    def passthroughs(self) -> dai.Node.OutputMap:
        return self._nn.passthroughs

    def run(self) -> None:
        """Methods inherited from ThreadedHostNode.

        Method runs with start of the pipeline.
        """
        self._checkNNArchive()

    def _checkNNArchive(self) -> None:
        if self._nn_archive is None:
            raise RuntimeError(
                f"NNArchive is not set. Use {self.setNNArchive.__name__} or {self.build.__name__} method to set it."
            )

    def getOutput(self, head: int) -> dai.Node.Output:
        """Obtains output of a parser for specified NeuralNetwork model head."""
        if head not in self._parsers:
            raise KeyError(
                f"Head {head} is not available. Available heads for the model {self._getModelName()} are {list(self._parsers.keys())}."
            )
        return self._parsers[head].out

    def _getModelName(self) -> str:
        return self._getConfig().model.metadata.name

    def _getConfig(self):
        return self._nn_archive.getConfig()



import cv2
import numpy as np


def calc_warp_corners_and_matches(
    ref_points: np.ndarray,
    dst_points: np.ndarray,
    image1: np.ndarray,
):
    H, mask = cv2.findHomography(
        ref_points,
        dst_points,
        cv2.USAC_MAGSAC,
        13.5,
        maxIters=1_000,
        confidence=0.8,
    )
    mask = mask.flatten()

    h, w = image1.shape[:2]
    corners_image1 = np.array(
        [[0, 0], [w - 1, 0], [w - 1, h - 1], [0, h - 1]], dtype=np.float32
    ).reshape(-1, 1, 2)

    warped_corners = cv2.perspectiveTransform(corners_image1, H)

    keypoints1 = [cv2.KeyPoint(p[0], p[1], 5) for p in ref_points]
    keypoints2 = [cv2.KeyPoint(p[0], p[1], 5) for p in dst_points]
    matches = [cv2.DMatch(i, i, 0) for i in range(len(mask)) if mask[i]]

    return warped_corners, [keypoints1, keypoints2, matches]


def xfeat_visualizer(image1, image2, features, draw_warp_corners=True):
    n = len(features)
    if n < 50:
        return image2

    ref_feats = features[0:n:2]
    tgt_feats = features[1:n:2]
    mkpts0 = np.array([[x.position.x, x.position.y] for x in ref_feats])
    mkpts1 = np.array([[x.position.x, x.position.y] for x in tgt_feats])

    result = calc_warp_corners_and_matches(
        mkpts0,
        mkpts1,
        image1,
    )

    warped_corners = result[0]
    keypoints1 = result[1][0]
    keypoints2 = result[1][1]
    matches = result[1][2]

    image2_with_corners = image2.copy()
    if draw_warp_corners:
        for i in range(len(warped_corners)):
            start_point = tuple(warped_corners[i - 1][0].astype(int))
            end_point = tuple(warped_corners[i][0].astype(int))
            cv2.line(image2_with_corners, start_point, end_point, (0, 255, 0), 4)

    debug_image = cv2.drawMatches(
        image1,
        keypoints1,
        image2_with_corners,
        keypoints2,
        matches,
        None,
        matchColor=(0, 255, 0),
        flags=2,
    )

    return debug_image



import depthai as dai
import numpy as np
from numpy.typing import NDArray


class Map2D(dai.Buffer):
    """Map2D class for storing a 2D map of floats.

    Attributes
    ----------
    map : NDArray[np.float32]
        2D map.
    width : int
        2D Map width.
    height : int
        2D Map height.
    """

    def __init__(self):
        """Initializes the Map2D object."""
        super().__init__()
        self._map: NDArray[np.float32] = np.array([])
        self._width: int = None
        self._height: int = None

    @property
    def map(self) -> NDArray[np.float32]:
        """Returns the 2D map.

        @return: 2D map.
        @rtype: NDArray[np.float32]
        """
        return self._map

    @map.setter
    def map(self, value: np.ndarray):
        """Sets the 2D map.

        @param value: 2D map.
        @type value: NDArray[np.float32]
        @raise TypeError: If value is not a numpy array.
        @raise ValueError: If value is not a 2D numpy array.
        @raise ValueError: If each element is not of type float.
        """
        if not isinstance(value, np.ndarray):
            raise TypeError(
                f"2D map must be of type np.ndarray, instead got {type(value)}."
            )
        if value.ndim != 2:
            raise ValueError("2D map must be a 2D array")
        if value.dtype != np.float32:
            raise ValueError("2D map must be an array of floats")
        self._map = value
        self._width = value.shape[1]
        self._height = value.shape[0]

    @property
    def width(self) -> int:
        """Returns the 2D map width.

        @return: 2D map width.
        @rtype: int
        """
        return self._width

    @property
    def height(self) -> int:
        """Returns the 2D map height.

        @return: 2D map height.
        @rtype: int
        """
        return self._height



import depthai as dai
import numpy as np
import pytest

from depthai_nodes.ml.messages.creators.image import create_image_message


def test_wrong_shape():
    img = np.random.random((10, 1, 10))
    with pytest.raises(
        ValueError, match="Unexpected image shape. Expected CHW or HWC, got"
    ):
        create_image_message(img)


def test_grayscale():
    img = (np.random.random((1, 100, 100)) * 255).astype(np.uint8)
    img_frame = create_image_message(img)

    assert img_frame.getType() == dai.ImgFrame.Type.GRAY8
    assert img_frame.getWidth() == 100
    assert img_frame.getHeight() == 100
    assert np.all(np.isclose(img_frame.getFrame(), img[0, :, :]))


def test_float_array():
    img = np.array([[[0.5, 0.5, 0.5]]])
    with pytest.raises(
        ValueError, match="Expected int type, got <class 'numpy.float64'>."
    ):
        create_image_message(img)


def test_bgr():
    img = (np.random.random((3, 100, 100)) * 255).astype(np.uint8)
    img_frame = create_image_message(img, False)

    assert img_frame.getType() == dai.ImgFrame.Type.BGR888i
    assert img_frame.getWidth() == 100
    assert img_frame.getHeight() == 100
    img = np.transpose(img, (1, 2, 0))
    img_r = img[:, :, 0]
    img_g = img[:, :, 1]
    img_b = img[:, :, 2]
    img = np.stack([img_b, img_g, img_r], axis=2)

    assert np.all(np.isclose(img_frame.getFrame(), img))


if __name__ == "__main__":
    pytest.main()



import depthai as dai
import numpy as np
from numpy.typing import NDArray


class SegmentationMask(dai.Buffer):
    """SegmentationMask class for a single- or multi-object segmentation mask.
    Background is represented with "-1" and foreground classes with non-negative
    integers.

    Attributes
    ----------
    mask: NDArray[np.int8]
        Segmentation mask.
    """

    def __init__(self):
        """Initializes the SegmentationMask object."""
        super().__init__()
        self._mask: NDArray[np.int8] = np.array([])

    @property
    def mask(self) -> NDArray[np.int8]:
        """Returns the segmentation mask.

        @return: Segmentation mask.
        @rtype: NDArray[np.int8]
        """
        return self._mask

    @mask.setter
    def mask(self, value: NDArray[np.int8]):
        """Sets the segmentation mask.

        @param value: Segmentation mask.
        @type value: NDArray[np.int8])
        @raise TypeError: If value is not a numpy array.
        @raise ValueError: If value is not a 2D numpy array.
        @raise ValueError: If each element is not of type int8.
        @raise ValueError: If each element is larger or equal to -1.
        """
        if not isinstance(value, np.ndarray):
            raise TypeError("Mask must be a numpy array.")
        if value.ndim != 2:
            raise ValueError("Mask must be 2D.")
        if value.dtype != np.int8:
            raise ValueError("Mask must be an array of int8.")
        if np.any((value < -1)):
            raise ValueError("Mask must be an array values larger or equal to -1.")
        self._mask = value


class SegmentationMasksSAM(dai.Buffer):
    """SegmentationMasksSAM class for storing segmentation masks.
    TODO: remove this message and use SegmentationMasks instead

    Attributes
    ----------
    masks: np.ndarray
        Mask coefficients.
    """

    def __init__(self):
        """Initializes the SegmentationMasks object."""
        super().__init__()
        self._masks: np.ndarray = np.array([])

    @property
    def masks(self) -> np.ndarray:
        """Returns the masks coefficients.

        @return: Masks coefficients.
        @rtype: np.ndarray
        """
        return self._masks

    @masks.setter
    def masks(self, value: np.ndarray):
        """Sets the masks coefficients.

        @param value: Masks coefficients.
        @type value: np.ndarray
        @raise TypeError: If the masks is not a numpy array.
        @raise ValueError: If the masks is not 3D.
        """
        if not isinstance(value, np.ndarray):
            raise TypeError("Masks must be a numpy array.")
        if value.ndim != 3:
            raise ValueError("Masks must be 3D.")
        self._masks = value



import cv2
import depthai as dai

from .utils.message_parsers import parse_image_message


def visualize_image(frame: dai.ImgFrame, message: dai.ImgFrame, extraParams: dict):
    """Visualizes the image on the frame."""
    image = parse_image_message(message)
    cv2.imshow("Image", image)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False



import numpy as np
import pytest

from depthai_nodes.ml.messages import Classifications
from depthai_nodes.ml.messages.creators.classification import (
    create_classification_message,
)


def test_none_classe():
    with pytest.raises(ValueError):
        create_classification_message(None, [0.5, 0.2, 0.3])


def test_none_scores():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], None)


def test_none_both():
    with pytest.raises(ValueError):
        create_classification_message(None, None)


def test_empty_classes():
    with pytest.raises(ValueError):
        create_classification_message([], [0.5, 0.2, 0.3])


def test_empty_scores():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], [])


def test_non_list_classes():
    with pytest.raises(ValueError):
        create_classification_message("cat", [0.5, 0.2, 0.3])


def test_non_np_scores():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], [int(1), int(2), int(3)])


def test_tuple_scores():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], (0.5, 0.2, 0.3))


def test_nd_scores():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], [[0.5, 0.2, 0.3]])


def test_mixed_scores():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], [0.5, 0.2, "30"])


def test_non_probability_scores():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], [0.2, 0.3, 0.1])


def test_non_probability_scores_2():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], [0.5, 0.5, 0.5])


def test_sum_above_upper_thr():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], [0.5, 0.21, 0.4])


def test_sum_below_upper_thr():
    create_classification_message(["cat", "dog", "bird"], [0.5, 0.11001000, 0.4])


def test_sum_above_lower_thr():
    create_classification_message(["cat", "dog", "bird"], [0.5, 0.2, 0.3])


def test_sum_below_bottom_thr():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], [0.5, 0.9, 0.3])


def test_mismatch_lengths():
    with pytest.raises(ValueError):
        create_classification_message(["cat", "dog", "bird"], [0.5, 0.2, 0.3, 0.4])


def test_correct_input():
    classes = ["cat", "dog", "bird"]
    scores = [0.2, 0.5, 0.3]

    correct_classes = ["dog", "bird", "cat"]
    correct_scores = [0.5, 0.3, 0.2]

    message = create_classification_message(classes, scores)

    assert isinstance(message, Classifications)
    assert message.classes == correct_classes
    assert np.all(message.scores == np.array(correct_scores, dtype=np.float32))
    assert isinstance(message.classes, list)
    assert isinstance(message.scores, np.ndarray)


def test_single_class_and_score():
    classes = ["cat"]
    scores = [1.0]

    message = create_classification_message(classes, scores)
    assert message.classes == ["cat"]
    assert message.scores == [1.0]


def test_very_small_scores():
    classes = ["cat", "dog", "bird"]
    scores = [1e-10, 1e-10, 1 - 2e-10]

    message = create_classification_message(classes, scores)

    assert isinstance(message, Classifications)
    assert message.classes == ["bird", "cat", "dog"]
    assert np.all(
        message.scores == np.array([1 - 2e-10, 1e-10, 1e-10], dtype=np.float32)
    )


def test_identical_scores():
    classes = ["cat", "dog", "bird"]
    scores = [1 / 3, 1 / 3, 1 / 3]

    message = create_classification_message(classes, scores)

    assert message.classes == classes
    assert np.all(message.scores == np.array(scores, dtype=np.float32))


def test_duplicate_scores():
    classes = ["cat", "dog", "bird"]
    scores = [0.4, 0.2, 0.4]

    correct_classes = ["cat", "bird", "dog"]
    correct_scores = [0.4, 0.4, 0.2]

    message = create_classification_message(classes, scores)

    assert message.classes == correct_classes
    assert np.all(message.scores == np.array(correct_scores, dtype=np.float32))


if __name__ == "__main__":
    pytest.main()



from typing import List, Tuple

import depthai as dai

from depthai_nodes.ml.parsers import *


def get_model_from_hub(model_slug: str, model_version_slug: str) -> dai.NNArchive:
    """Get the model from the HubAI and return the NN archive."""
    print(
        f"Downloading model {model_slug} with version {model_version_slug} from HubAI..."
    )
    modelDescription = dai.NNModelDescription(
        modelSlug=model_slug, modelVersionSlug=model_version_slug, platform="RVC2"
    )
    archivePath = dai.getModelFromZoo(modelDescription)
    print("Download successful!")
    nn_archive = dai.NNArchive(archivePath)

    return nn_archive


def get_parser_from_archive(nn_archive: dai.NNArchive) -> str:
    """Get the required parser from the NN archive."""
    try:
        required_parser = nn_archive.getConfig().model.heads[0].parser
    except AttributeError:
        print(
            "This NN archive does not have a parser. Please use NN archives that have parsers."
        )
        exit(1)

    print(f"Required parser: {required_parser}")

    return required_parser


def get_parser(nn_archive: dai.NNArchive) -> Tuple[dai.ThreadedNode, str]:
    """Map the parser from the NN archive to the actual parser in depthai-nodes."""
    required_parser = get_parser_from_archive(nn_archive)

    if required_parser == "YOLO" or required_parser == "SSD":
        return None, required_parser

    parser = globals().get(required_parser, None)

    if parser is None:
        raise NameError(
            f"Parser {required_parser} is not available in the depthai_nodes.ml.parsers module."
        )

    return parser, required_parser


def get_inputs_from_archive(nn_archive: dai.NNArchive) -> List:
    """Get all inputs from NN archive."""
    try:
        inputs = nn_archive.getConfig().model.inputs
    except AttributeError:
        print(
            "This NN archive does not have an input shape. Please use NN archives that have input shapes."
        )
        exit(1)

    return inputs


def get_input_shape(nn_archive: dai.NNArchive) -> Tuple[int, int]:
    """Get the input shape of the model from the NN archive."""
    inputs = get_inputs_from_archive(nn_archive)

    if len(inputs) > 1:
        raise ValueError(
            "This model has more than one input. Currently, only models with one input are supported."
        )

    try:
        input_shape = inputs[0].shape[2:][::-1]
    except AttributeError:
        print(
            "This NN archive does not have an input shape. Please use NN archives that have input shapes."
        )
        exit(1)

    print(f"Input shape: {input_shape}")

    return input_shape



from typing import Dict

import depthai as dai

from .ml.parsers import *
from .ml.parsers.base_parser import BaseParser
from .ml.parsers.utils import decode_head


class ParserGenerator(dai.node.ThreadedHostNode):
    """General interface for instantiating parsers based on the provided model archive.

    The `build` method creates parsers based on the head information stored in the NN Archive. The method then returns a dictionary of these parsers.
    """

    def __init__(self):
        super().__init__()

    def build(self, nn_archive: dai.NNArchive, head_index: int = None) -> Dict:
        """Instantiates parsers based on the provided model archive.

        @param nn_archive: NN Archive of the model.
        @type nn_archive: dai.NNArchive
        @param head_index: Index of the head to be used for parsing. If not provided,
            each head will instantiate a separate parser.
        @type head_index: int
        @return: A dictionary of instantiated parsers.
        @rtype: Dict[int, BaseParser]
        """

        heads = nn_archive.getConfig().model.heads
        indexes = range(len(heads))

        if len(heads) == 0:
            raise ValueError("No heads defined in the NN Archive.")

        if head_index:
            heads = [heads[head_index]]
            indexes = [head_index]

        parsers = {}
        pipeline = self.getParentPipeline()

        for index, head in zip(indexes, heads):
            parser_name = head.parser
            parser = globals().get(parser_name)

            if parser is None:
                raise ValueError(f"Parser {parser_name} not a valid parser class.")
            elif not issubclass(parser, BaseParser):
                raise ValueError(
                    f"Parser {parser_name} does not inherit from BaseParser class."
                )

            head = decode_head(head)
            parsers[index] = pipeline.create(parser).build(head)

        return parsers

    def run(self):
        pass



import cv2
import depthai as dai
import numpy as np

from depthai_nodes.ml.messages import SegmentationMasksSAM

from .utils.colors import get_adas_colors, get_ewasr_colors, get_selfie_colors
from .utils.message_parsers import parse_fast_sam_message, parse_segmentation_message


def visualize_segmentation(
    frame: dai.ImgFrame, message: dai.ImgFrame, extraParams: dict
):
    mask = parse_segmentation_message(message)
    frame = cv2.resize(frame, (mask.shape[1], mask.shape[0]))

    n_classes = extraParams.get("n_classes", None)

    if n_classes is None:
        raise ValueError("Number of classes not provided in NN archive metadata.")

    if n_classes == 2:
        COLORS = get_selfie_colors()
    elif n_classes == 3:
        COLORS = get_ewasr_colors()
    else:
        COLORS = get_adas_colors()

    colormap = np.zeros((mask.shape[0], mask.shape[1], 3), dtype=np.uint8)
    for class_id, color in enumerate(COLORS):
        m = mask == class_id
        colormap[m] = color

    alpha = 0.5
    overlay = cv2.addWeighted(colormap, alpha, frame, 1 - alpha, 0)

    cv2.imshow("Segmentation", overlay)

    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False


def _fastsam_show_masks(
    annotation,
    image,
):
    if len(annotation) == 0:
        return image
    n, h, w = annotation.shape  # batch, height, width
    areas = np.sum(annotation, axis=(1, 2))
    annotation = annotation[np.argsort(areas)]

    index = (annotation != 0).argmax(axis=0)
    color = np.random.random((n, 1, 1, 3)) * 255
    transparency = np.ones((n, 1, 1, 1)) * 0.6
    visual = np.concatenate([color, transparency], axis=-1)
    mask_image = np.expand_dims(annotation, -1) * visual

    show = np.zeros((h, w, 4), dtype=np.float32)
    h_indices, w_indices = np.meshgrid(np.arange(h), np.arange(w), indexing="ij")
    indices = (index[h_indices, w_indices], h_indices, w_indices, slice(None))
    show[h_indices, w_indices, :] = mask_image[indices]

    overlay = image.astype(np.float32)
    alpha = show[:, :, 3:4]
    overlay[:, :, :3] = overlay[:, :, :3] * (1 - alpha) + show[:, :, :3] * alpha

    return overlay.astype(np.uint8)


def visualize_fastsam(
    frame: dai.ImgFrame, message: SegmentationMasksSAM, extraParams: dict
):
    masks = parse_fast_sam_message(message)

    if masks is not None:
        for i, mask in enumerate(masks):
            masks[i] = cv2.morphologyEx(
                mask.astype(np.uint8), cv2.MORPH_OPEN, np.ones((8, 8), np.uint8)
            )

        overlay = _fastsam_show_masks(masks, frame)
    else:
        overlay = frame

    cv2.imshow("FastSAM segmentation", overlay)

    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False



import re

import numpy as np
import pytest

from depthai_nodes.ml.messages import Map2D
from depthai_nodes.ml.messages.creators.map import create_map_message

np.random.seed(0)


def test_not_numpy_array():
    with pytest.raises(ValueError, match="Expected numpy array, got <class 'list'>."):
        create_map_message([1, 2, 3])


def test_not_2d_or_3d_input():
    with pytest.raises(ValueError, match="Expected 2D or 3D input, got 1D input."):
        create_map_message(np.array([1, 2, 3]))


def test_wrong_input_shape():
    with pytest.raises(
        ValueError,
        match=re.escape("Unexpected map shape. Expected NHW or HWN, got (3, 1, 3)."),
    ):
        create_map_message(np.array([[[1, 2, 3]], [[4, 5, 6]], [[7, 8, 9]]]))


def test_map_with_scaling():
    map = np.random.rand(320, 640, 1)
    map[0, 0, 0] = 0.0  # ensure that min==0
    map[1, 0, 0] = 1.0  # ensure that max==1
    unscaled_map = map * 10

    message = create_map_message(map=unscaled_map, min_max_scaling=True)

    assert isinstance(message, Map2D)
    assert message.width == 640
    assert message.height == 320

    scaled_map = message.map
    assert scaled_map.shape == map[:, :, 0].shape
    assert np.all(np.isclose(map[:, :, 0], scaled_map))


def test_map_without_scaling():
    map = np.random.rand(320, 640, 1)
    unscaled_map = map * 10

    message = create_map_message(map=unscaled_map, min_max_scaling=False)

    assert isinstance(message, Map2D)
    assert message.width == 640
    assert message.height == 320

    unscaled_map2 = message.map
    assert unscaled_map2.shape == unscaled_map[:, :, 0].shape
    assert np.all(np.isclose(unscaled_map[:, :, 0], unscaled_map2))


if __name__ == "__main__":
    pytest.main()



from typing import List, Union

import numpy as np

from ...messages import Keypoint, Keypoints


def create_keypoints_message(
    keypoints: Union[np.ndarray, List[List[float]]],
    scores: Union[np.ndarray, List[float]] = None,
    confidence_threshold: float = None,
) -> Keypoints:
    """Create a DepthAI message for the keypoints.

    @param keypoints: Detected 2D or 3D keypoints of shape (N,2 or 3) meaning [...,[x, y],...] or [...,[x, y, z],...].
    @type keypoints: np.ndarray or List[List[float]]
    @param scores: Confidence scores of the detected keypoints.
    @type scores: Optional[np.ndarray or List[float]]
    @param confidence_threshold: Confidence threshold of keypoint detections.
    @type confidence_threshold: Optional[float]

    @return: Keypoints message containing the detected keypoints.
    @rtype: Keypoints

    @raise ValueError: If the keypoints are not a numpy array or list.
    @raise ValueError: If the scores are not a numpy array or list.
    @raise ValueError: If scores and keypoints do not have the same length.
    @raise ValueError: If score values are not floats.
    @raise ValueError: If score values are not between 0 and 1.
    @raise ValueError: If the confidence threshold is not a float.
    @raise ValueError: If the confidence threshold is not between 0 and 1.
    @raise ValueError: If the keypoints are not of shape (N,2 or 3).
    @raise ValueError: If the keypoints 2nd dimension is not of size E{2} or E{3}.
    """

    if not isinstance(keypoints, (np.ndarray, list)):
        raise ValueError(
            f"Keypoints should be numpy array or list, got {type(keypoints)}."
        )

    if scores is not None:
        if not isinstance(scores, (np.ndarray, list)):
            raise ValueError(
                f"Scores should be numpy array or list, got {type(scores)}."
            )

        if len(keypoints) != len(scores):
            raise ValueError(
                "Keypoints and scores should have the same length. Got {} keypoints and {} scores.".format(
                    len(keypoints), len(scores)
                )
            )

        if not all(isinstance(score, (float, np.floating)) for score in scores):
            raise ValueError("Scores should only contain float values.")
        if not all(0 <= score <= 1 for score in scores):
            raise ValueError("Scores should only contain values between 0 and 1.")

    if confidence_threshold is not None:
        if not isinstance(confidence_threshold, float):
            raise ValueError(
                f"The confidence_threshold should be float, got {type(confidence_threshold)}."
            )

        if not (0 <= confidence_threshold <= 1):
            raise ValueError(
                f"The confidence_threshold should be between 0 and 1, got confidence_threshold {confidence_threshold}."
            )

    if len(keypoints) != 0:
        dimension = len(keypoints[0])
        if dimension != 2 and dimension != 3:
            raise ValueError(
                f"All keypoints should be of dimension 2 or 3, got dimension {dimension}."
            )

    if isinstance(keypoints, list):
        for keypoint in keypoints:
            if not isinstance(keypoint, list):
                raise ValueError(
                    f"Keypoints should be list of lists or np.array, got list of {type(keypoint)}."
                )
            if len(keypoint) != dimension:
                raise ValueError(
                    "All keypoints have to be of same dimension e.g. [x, y] or [x, y, z], got mixed inner dimensions."
                )
            for coord in keypoint:
                if not isinstance(coord, (float, np.floating)):
                    raise ValueError(
                        f"Keypoints inner list should contain only float, got {type(coord)}."
                    )

    keypoints = np.array(keypoints)
    if scores is not None:
        scores = np.array(scores)

    if len(keypoints) != 0:
        if len(keypoints.shape) != 2:
            raise ValueError(
                f"Keypoints should be of shape (N,2 or 3) got {keypoints.shape}."
            )

        use_3d = keypoints.shape[1] == 3

    keypoints_msg = Keypoints()
    points = []
    for i, keypoint in enumerate(keypoints):
        if scores is not None and confidence_threshold is not None:
            if scores[i] < confidence_threshold:
                continue
        pt = Keypoint()
        pt.x = float(keypoint[0])
        pt.y = float(keypoint[1])
        pt.z = float(keypoint[2]) if use_3d else 0.0
        if scores is not None:
            pt.confidence = float(scores[i])
        points.append(pt)

    keypoints_msg.keypoints = points
    return keypoints_msg



from typing import Any, Dict, List, Optional

import cv2
import depthai as dai
import numpy as np

from ..messages.creators import create_detection_message
from .base_parser import BaseParser
from .utils.bbox_format_converters import normalize_bboxes, xyxy_to_xywh
from .utils.masks_utils import (
    get_segmentation_outputs,
    process_single_mask,
    reshape_seg_outputs,
)
from .utils.yolo import (
    YOLOSubtype,
    decode_yolo_output,
    parse_kpts,
)


class YOLOExtendedParser(BaseParser):
    """Parser class for parsing the output of the YOLO Instance Segmentation and Pose
    Estimation models.

    Attributes
    ----------
    conf_threshold : float
        Confidence score threshold for detected faces.
    n_classes : int
        Number of classes in the model.
    iou_threshold : float
        Intersection over union threshold.
    mask_conf : float
        Mask confidence threshold.
    n_keypoints : int
        Number of keypoints in the model.
    anchors : Optional[List[np.ndarray]]
        Anchors for the YOLO model (optional).
    subtype : str
        Version of the YOLO model.


    Output Message/s
    ----------------
    **Type**: ImgDetectionsExtended

    **Description**: Message containing bounding boxes, labels, confidence scores, and keypoints or masks and protos of the detected objects.
    """

    _DET_MODE = 0
    _KPTS_MODE = 1
    _SEG_MODE = 2

    def __init__(
        self,
        conf_threshold: float = 0.5,
        n_classes: int = 1,
        iou_threshold: float = 0.5,
        mask_conf: float = 0.5,
        n_keypoints: int = 17,
        anchors: Optional[List[np.ndarray]] = None,
        subtype: str = "",
    ):
        """Initialize the parser node.

        @param conf_threshold: The confidence threshold for the detections
        @type conf_threshold: float
        @param n_classes: The number of classes in the model
        @type n_classes: int
        @param iou_threshold: The intersection over union threshold
        @type iou_threshold: float
        @param mask_conf: The mask confidence threshold
        @type mask_conf: float
        @param n_keypoints: The number of keypoints in the model
        @type n_keypoints: int
        @param anchors: The anchors for the YOLO model
        @type anchors: Optional[List[np.ndarray]]
        @param subtype: The version of the YOLO model
        @type subtype: str
        """
        super().__init__()

        self.output_layer_names = []
        self.conf_threshold = conf_threshold
        self.n_classes = n_classes
        self.iou_threshold = iou_threshold
        self.mask_conf = mask_conf
        self.n_keypoints = n_keypoints
        self.anchors = anchors
        try:
            self.subtype = YOLOSubtype(subtype.lower())
        except ValueError as err:
            raise ValueError(
                f"Invalid YOLO version {subtype}. Supported YOLO versions are {[e.value for e in YOLOSubtype][:-1]}."
            ) from err

    def setOutputLayerNames(self, output_layer_names: List[str]) -> None:
        """Sets the output layer names for the parser.

        @param output_layer_names: The output layer names for the parser.
        @type output_layer_names: List[str]
        """
        self.output_layer_names = output_layer_names

    def setConfidenceThreshold(self, threshold: float) -> None:
        """Sets the confidence score threshold for detected faces.

        @param threshold: Confidence score threshold for detected faces.
        @type threshold: float
        """
        self.conf_threshold = threshold

    def setNumClasses(self, n_classes: int) -> None:
        """Sets the number of classes in the model.

        @param numClasses: The number of classes in the model.
        @type numClasses: int
        """
        self.n_classes = n_classes

    def setIouThreshold(self, iou_threshold: float) -> None:
        """Sets the intersection over union threshold.

        @param iou_threshold: The intersection over union threshold.
        @type iou_threshold: float
        """
        self.iou_threshold = iou_threshold

    def setMaskConfidence(self, mask_conf: float) -> None:
        """Sets the mask confidence threshold.

        @param mask_conf: The mask confidence threshold.
        @type mask_conf: float
        """
        self.mask_conf = mask_conf

    def setNumKeypoints(self, n_keypoints: int) -> None:
        """Sets the number of keypoints in the model.

        @param n_keypoints: The number of keypoints in the model.
        @type n_keypoints: int
        """
        self.n_keypoints = n_keypoints

    def setAnchors(self, anchors: List[np.ndarray]) -> None:
        """Sets the anchors for the YOLO model.

        @param anchors: The anchors for the YOLO model.
        @type anchors: List[np.ndarray]
        """
        self.anchors = anchors

    def setSubtype(self, subtype: str) -> None:
        """Sets the subtype of the YOLO model.

        @param subtype: The subtype of the YOLO model.
        @type subtype: YOLOSubtype
        """
        try:
            self.subtype = YOLOSubtype(subtype.lower())
        except ValueError as err:
            raise ValueError(
                f"Invalid YOLO subtype {subtype}. Supported YOLO subtypes are {[e.value for e in YOLOSubtype][:-1]}."
            ) from err

    def build(
        self,
        head_config: Dict[str, Any],
    ):
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: YOLOExtendedParser
        """

        output_layers = head_config.get("outputs", [])
        bbox_layer_names = [name for name in output_layers if "_yolo" in name]
        kps_layer_names = [name for name in output_layers if "kpt_output" in name]
        masks_layer_names = [name for name in output_layers if "_masks" in name]

        if not bbox_layer_names:
            raise ValueError(
                "YOLOExtendedParser requires the output layers of the bounding boxes detection head."
            )
        if (kps_layer_names and masks_layer_names) or (
            not bbox_layer_names and (kps_layer_names or masks_layer_names)
        ):
            raise ValueError(
                "YOLOExtendedParser requires either the output layers of the keypoints detection head or the output layers of the masks detection head along with the output layers of the bounding boxes detection head but not both."
            )

        self.output_layer_names = bbox_layer_names + kps_layer_names + masks_layer_names

        self.conf_threshold = head_config.get("conf_threshold", self.conf_threshold)
        self.n_classes = head_config.get("n_classes", self.n_classes)
        self.iou_threshold = head_config.get("iou_threshold", self.iou_threshold)
        self.mask_conf = head_config.get("mask_conf", self.mask_conf)
        self.anchors = head_config.get("anchors", self.anchors)
        self.n_keypoints = head_config.get("n_keypoints", self.n_keypoints)
        subtype = head_config.get("subtype", self.subtype)
        try:
            self.subtype = YOLOSubtype(subtype.lower())
        except ValueError as err:
            raise ValueError(
                f"Invalid YOLO subtype {subtype}. Supported YOLO subtypes are {[e.value for e in YOLOSubtype][:-1]}."
            ) from err

        return self

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped, no more data
            # Get all the layer names
            layer_names = self.output_layer_names or output.getAllLayerNames()

            outputs_names = sorted(
                [name for name in layer_names if "_yolo" in name or "yolo-" in name]
            )
            outputs_values = [
                output.getTensor(o, dequantize=True).astype(np.float32)
                for o in outputs_names
            ]

            if (
                any("kpt_output" in name for name in layer_names)
                and self.subtype != YOLOSubtype.P
            ):
                mode = self._KPTS_MODE
                # Get the keypoint outputs
                kpts_output_names = sorted(
                    [name for name in layer_names if "kpt_output" in name]
                )
                kpts_outputs = [
                    output.getTensor(o, dequantize=True).astype(np.float32)
                    for o in kpts_output_names
                ]
            elif (
                any("_masks" in name for name in layer_names)
                and self.subtype != YOLOSubtype.P
            ):
                mode = self._SEG_MODE
                # Get the segmentation outputs
                (
                    masks_outputs_values,
                    protos_output,
                    protos_len,
                ) = get_segmentation_outputs(output)
            else:
                mode = self._DET_MODE

            if (
                len(outputs_values[0].shape) == 4
                and outputs_values[0].shape[-1] == outputs_values[1].shape[-1]
            ):
                # RVC4
                outputs_values = [o.transpose((0, 3, 1, 2)) for o in outputs_values]
                if mode == self._SEG_MODE:
                    (
                        protos_output,
                        protos_len,
                        masks_outputs_values,
                    ) = reshape_seg_outputs(
                        protos_output, protos_len, masks_outputs_values
                    )

            # Get the model's input shape
            strides = (
                [8, 16, 32]
                if self.subtype
                not in [YOLOSubtype.V3UT, YOLOSubtype.V3T, YOLOSubtype.V4T]
                else [16, 32]
            )
            input_shape = tuple(
                dim * strides[0] for dim in outputs_values[0].shape[2:4]
            )

            # Decode the outputs
            results = decode_yolo_output(
                outputs_values,
                strides,
                self.anchors,
                kpts=kpts_outputs if mode == self._KPTS_MODE else None,
                conf_thres=self.conf_threshold,
                iou_thres=self.iou_threshold,
                num_classes=self.n_classes,
                det_mode=mode == self._DET_MODE,
                subtype=self.subtype,
            )

            bboxes, labels, scores, additional_output = [], [], [], []
            final_mask = np.full(input_shape, -1, dtype=float)
            for i in range(results.shape[0]):
                bbox, conf, label, other = (
                    results[i, :4],
                    results[i, 4],
                    results[i, 5].astype(int),
                    results[i, 6:],
                )

                bbox = xyxy_to_xywh(bbox.reshape(1, 4))
                bbox = normalize_bboxes(
                    bbox, height=input_shape[0], width=input_shape[1]
                )[0]
                bboxes.append(bbox)
                labels.append(int(label))
                scores.append(conf)

                if mode == self._KPTS_MODE:
                    kpts = parse_kpts(other, self.n_keypoints, input_shape)
                    additional_output.append(kpts)
                elif mode == self._SEG_MODE:
                    seg_coeff = other.astype(int)
                    hi, ai, xi, yi = seg_coeff
                    mask_coeff = masks_outputs_values[hi][
                        0, ai * protos_len : (ai + 1) * protos_len, yi, xi
                    ]
                    mask = process_single_mask(
                        protos_output[0], mask_coeff, self.mask_conf, bbox
                    )
                    # Resize mask to input shape
                    resized_mask = cv2.resize(
                        mask,
                        (input_shape[1], input_shape[0]),
                        interpolation=cv2.INTER_NEAREST,
                    )
                    # Fill the final mask with the instance values
                    final_mask[resized_mask > 0] = i

            bboxes = np.array(bboxes)
            bboxes = np.clip(bboxes, 0, 1)

            if mode == self._KPTS_MODE:
                additional_output = np.array(additional_output)
                keypoints = np.array([])
                keypoints_scores = np.array([])
                if additional_output.size > 0:
                    keypoints = additional_output[:, :, :2]
                    keypoints_scores = additional_output[:, :, 2]

                keypoints = np.clip(keypoints, 0, 1)
                detections_message = create_detection_message(
                    bboxes=bboxes,
                    scores=np.array(scores),
                    labels=np.array(labels),
                    keypoints=keypoints,
                    keypoints_scores=keypoints_scores,
                )
            elif mode == self._SEG_MODE:
                detections_message = create_detection_message(
                    bboxes=bboxes,
                    scores=np.array(scores),
                    labels=np.array(labels),
                    masks=final_mask,
                )
            else:
                detections_message = create_detection_message(
                    bboxes=bboxes,
                    scores=np.array(scores),
                    labels=np.array(labels),
                )

            detections_message.setTimestamp(output.getTimestamp())

            self.out.send(detections_message)



from typing import List

import cv2
import depthai as dai
from visualization.visualizers import xfeat_visualizer

from depthai_nodes.ml.parsers import XFeatMonoParser, XFeatStereoParser


def xfeat_mono(nn_archive: dai.NNArchive, input_shape: List[int], fps_limit: int):
    """Run the XFeatMonoParser on a single camera.

    It lets you set the reference frame by pressing S-key.
    """
    previous_frame = None
    with dai.Pipeline() as pipeline:
        # Set up camera
        cam = pipeline.create(dai.node.Camera).build()

        # Set up the neural network
        network = pipeline.create(dai.node.NeuralNetwork).build(
            cam.requestOutput(
                input_shape, type=dai.ImgFrame.Type.BGR888p, fps=fps_limit
            ),
            nn_archive,
        )

        # Set up parser
        parser = XFeatMonoParser()
        parser.setOriginalSize(tuple(input_shape))
        parser.setInputSize(tuple(input_shape))
        parser.setMaxKeypoints(2048)

        # Linking
        network.out.link(parser.input)

        # Set up queue
        camera_queue = network.passthrough.createOutputQueue()
        parser_queue = parser.out.createOutputQueue()

        pipeline.start()

        while pipeline.isRunning():
            frame: dai.ImgFrame = camera_queue.get().getCvFrame()
            message: dai.TrackedFeatures = (
                parser_queue.get()
            )  # get message from the queue
            features = message.trackedFeatures
            if previous_frame is not None:
                resulting_frame = xfeat_visualizer(previous_frame, frame, features)
            else:
                resulting_frame = frame
            number_of_matches = len(features) // 2
            cv2.putText(
                resulting_frame,
                f"Number of matches: {number_of_matches}",
                (10, 30),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.5,
                (255, 255, 255),
                1,
            )
            cv2.imshow("XFeat", resulting_frame)

            key_pressed = cv2.waitKey(1)
            if key_pressed == ord("s"):
                parser.setTrigger()  # trigger to set the reference frame
                previous_frame = frame
            if key_pressed == ord("q"):
                cv2.destroyAllWindows()
                pipeline.stop()
                break


def xfeat_stereo(nn_archive: dai.NNArchive, input_shape: List[int], fps_limit: int):
    """Run the XFeatStereoParser on stereo cameras - left and right - and match the features."""
    with dai.Pipeline() as pipeline:
        device: dai.Device = pipeline.getDefaultDevice()
        available_cameras = [
            camera.name for camera in device.getConnectedCameraFeatures()
        ]

        if "left" not in available_cameras or "right" not in available_cameras:
            raise RuntimeError(
                f"Stereo cameras are not available! Available cameras: {available_cameras}"
            )

        left_cam = pipeline.create(dai.node.Camera).build(dai.CameraBoardSocket.CAM_B)
        right_cam = pipeline.create(dai.node.Camera).build(dai.CameraBoardSocket.CAM_C)

        left_network = pipeline.create(dai.node.NeuralNetwork).build(
            left_cam.requestOutput(
                input_shape, type=dai.ImgFrame.Type.RGB888p, fps=fps_limit
            ),
            nn_archive,
        )
        left_network.setNumInferenceThreads(2)

        right_network = pipeline.create(dai.node.NeuralNetwork).build(
            right_cam.requestOutput(
                input_shape, type=dai.ImgFrame.Type.RGB888p, fps=fps_limit
            ),
            nn_archive,
        )
        right_network.setNumInferenceThreads(2)

        parser = pipeline.create(XFeatStereoParser)
        parser.setOriginalSize(tuple(input_shape))
        parser.setInputSize(tuple(input_shape))
        parser.setMaxKeypoints(512)

        left_network.out.link(parser.reference_input)
        right_network.out.link(parser.target_input)

        left_cam_queue = left_network.passthrough.createOutputQueue()
        right_cam_queue = right_network.passthrough.createOutputQueue()
        parser_queue = parser.out.createOutputQueue()

        pipeline.start()

        while pipeline.isRunning():
            left_frame: dai.ImgFrame = left_cam_queue.get().getCvFrame()
            right_frame: dai.ImgFrame = right_cam_queue.get().getCvFrame()
            features: dai.TrackedFeatures = parser_queue.get()
            features = features.trackedFeatures

            resulting_frame = xfeat_visualizer(
                left_frame, right_frame, features, draw_warp_corners=False
            )
            number_of_matches = len(features) // 2
            cv2.putText(
                resulting_frame,
                f"Number of matches: {number_of_matches}",
                (10, 30),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.5,
                (255, 255, 255),
                1,
            )
            cv2.imshow("XFeat Stereo", resulting_frame)

            if cv2.waitKey(1) == ord("q"):
                cv2.destroyAllWindows()
                pipeline.stop()
                break



import cv2
import depthai as dai
import numpy as np

from depthai_nodes.ml.messages import (
    Clusters,
    ImgDetectionsExtended,
    Lines,
)

from .utils.message_parsers import (
    parse_cluster_message,
    parse_detection_message,
    parse_line_detection_message,
    parse_yolo_kpts_message,
)


def visualize_detections(
    frame: dai.ImgFrame, message: ImgDetectionsExtended, extraParams: dict
):
    """Visualizes the detections on the frame. Detections are given in xywh format
    (ImgDetectionsExtended).

    Also, checks if there are any keypoints available to visualize.
    """
    labels = extraParams.get("classes", None)
    detections = parse_detection_message(message)
    for detection in detections:
        x_center, y_center, width, height = (
            detection.x_center,
            detection.y_center,
            detection.width,
            detection.height,
        )
        xmin = x_center - width / 2
        ymin = y_center - height / 2
        xmax = x_center + width / 2
        ymax = y_center + height / 2

        if xmin > 1 or ymin > 1 or xmax > 1 or ymax > 1:
            xmin = int(xmin)
            ymin = int(ymin)
            xmax = int(xmax)
            ymax = int(ymax)
        else:
            xmin = int(xmin * frame.shape[1])
            ymin = int(ymin * frame.shape[0])
            xmax = int(xmax * frame.shape[1])
            ymax = int(ymax * frame.shape[0])

        if detection.angle == 0:
            cv2.rectangle(
                frame, (int(xmin), int(ymin)), (int(xmax), int(ymax)), (255, 0, 0), 2
            )
        else:
            # Rotated rectangle
            x_center = int(x_center * frame.shape[1])
            y_center = int(y_center * frame.shape[0])
            width = int(width * frame.shape[1])
            height = int(height * frame.shape[0])
            box = cv2.boxPoints(
                ((x_center, y_center), (width, height), detection.angle)
            )
            box = np.int0(box)
            cv2.drawContours(frame, [box], 0, (255, 0, 0), 2)

        try:
            keypoints = detection.keypoints
            for kp in keypoints:
                cv2.circle(
                    frame,
                    (int(kp.x * frame.shape[1]), int(kp.y * frame.shape[0])),
                    5,
                    (0, 0, 255),
                    -1,
                )
        except Exception:
            pass

        cv2.putText(
            frame,
            f"{detection.confidence * 100:.2f}%",
            (int(xmin) + 10, int(ymin) + 20),
            cv2.FONT_HERSHEY_TRIPLEX,
            0.5,
            255,
        )
        if labels is not None:
            cv2.putText(
                frame,
                labels[detection.label],
                (int(xmin) + 10, int(ymin) + 40),
                cv2.FONT_HERSHEY_TRIPLEX,
                0.5,
                255,
            )

    cv2.imshow("Detections", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False


def visualize_detections_xyxy(
    frame: dai.ImgFrame, message: dai.ImgDetections, extraParams: dict
):
    """Visualize the detections on the frame.

    The detections are in xyxy format (dai.ImgDetections).
    """

    labels = extraParams.get("classes", None)
    detections = parse_detection_message(message)
    for detection in detections:
        xmin, ymin, xmax, ymax = (
            detection.xmin,
            detection.ymin,
            detection.xmax,
            detection.ymax,
        )
        if xmin > 1 or ymin > 1 or xmax > 1 or ymax > 1:
            xmin = int(xmin)
            ymin = int(ymin)
            xmax = int(xmax)
            ymax = int(ymax)
        else:
            xmin = int(xmin * frame.shape[1])
            ymin = int(ymin * frame.shape[0])
            xmax = int(xmax * frame.shape[1])
            ymax = int(ymax * frame.shape[0])
        cv2.rectangle(
            frame, (int(xmin), int(ymin)), (int(xmax), int(ymax)), (255, 0, 0), 2
        )

        cv2.putText(
            frame,
            f"{detection.confidence * 100:.2f}%",
            (int(xmin) + 10, int(ymin) + 20),
            cv2.FONT_HERSHEY_TRIPLEX,
            0.5,
            255,
        )
        if labels is not None:
            cv2.putText(
                frame,
                labels[detection.label],
                (int(xmin) + 10, int(ymin) + 40),
                cv2.FONT_HERSHEY_TRIPLEX,
                0.5,
                255,
            )

    cv2.imshow("Detections", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False


def visualize_line_detections(frame: dai.ImgFrame, message: Lines, extraParams: dict):
    """Visualizes the lines on the frame."""
    lines = parse_line_detection_message(message)
    h, w = frame.shape[:2]
    for line in lines:
        x1 = line.start_point.x * w
        y1 = line.start_point.y * h
        x2 = line.end_point.x * w
        y2 = line.end_point.y * h
        cv2.line(frame, (int(x1), int(y1)), (int(x2), int(y2)), (0, 255, 0), 3, 16)

    cv2.putText(
        frame,
        f"Number of lines: {len(lines)}",
        (2, frame.shape[0] - 4),
        cv2.FONT_HERSHEY_COMPLEX,
        0.5,
        (255, 0, 0),
    )
    cv2.imshow("Lines", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False


def visualize_yolo_extended(
    frame: dai.ImgFrame, message: ImgDetectionsExtended, extraParams: dict
):
    """Visualizes the YOLO pose detections or instance segmentation on the frame."""
    detections = parse_yolo_kpts_message(message)

    classes = extraParams.get("classes", None)
    if classes is None:
        raise ValueError("Classes are required for visualization.")
    task = extraParams.get("n_keypoints", None)
    if task is None:
        task = "segmentation"
    else:
        task = "keypoints"

    for detection in detections:
        x_center, y_center, width, height = (
            detection.x_center,
            detection.y_center,
            detection.width,
            detection.height,
        )
        xmin = x_center - width / 2
        ymin = y_center - height / 2
        xmax = x_center + width / 2
        ymax = y_center + height / 2
        cv2.rectangle(
            frame,
            (int(xmin * frame.shape[1]), int(ymin * frame.shape[0])),
            (int(xmax * frame.shape[1]), int(ymax * frame.shape[0])),
            (255, 0, 0),
            2,
        )
        cv2.putText(
            frame,
            f"{detection.confidence * 100:.2f}%",
            (int(xmin * frame.shape[1]) + 10, int(ymin * frame.shape[0]) + 20),
            cv2.FONT_HERSHEY_TRIPLEX,
            0.5,
            255,
        )
        cv2.putText(
            frame,
            f"{classes[detection.label]}",
            (int(xmin * frame.shape[1]) + 10, int(ymin * frame.shape[0]) + 40),
            cv2.FONT_HERSHEY_TRIPLEX,
            0.5,
            255,
        )

        if task == "keypoints":
            keypoints = detection.keypoints
            for keypoint in keypoints:
                x, y, visibility = (
                    keypoint.x * frame.shape[1],
                    keypoint.y * frame.shape[0],
                    keypoint.confidence,
                )
                if visibility > 0.8:
                    cv2.circle(frame, (int(x), int(y)), 2, (0, 255, 0), -1)

    if task == "segmentation":
        mask = message.masks
        mask = mask > -0.5
        frame[mask] = frame[mask] * 0.5 + np.array((0, 255, 0)) * 0.5
    cv2.imshow("YOLO Pose Estimation", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False


def visualize_lane_detections(
    frame: dai.ImgFrame, message: Clusters, extraParams: dict
):
    """Visualizes the lines on the frame."""
    clusters = parse_cluster_message(message)
    h, w, _ = frame.shape
    for cluster in clusters:
        for point in cluster.points:
            x = int(point.x * w)
            y = int(point.y * h)
            cv2.circle(frame, (x, y), 3, (0, 255, 0), -1)

    # draw lines between points
    for cluster in clusters:
        for i in range(len(cluster.points) - 1):
            cv2.line(
                frame,
                (int(cluster.points[i].x * w), int(cluster.points[i].y * h)),
                (int(cluster.points[i + 1].x * w), int(cluster.points[i + 1].y * h)),
                (0, 255, 0),
                2,
            )

    cv2.imshow("Lane Detection", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False



import depthai as dai

from depthai_nodes.ml.messages import (
    Classifications,
    Clusters,
    ImgDetectionsExtended,
    Keypoints,
    Lines,
    Map2D,
    SegmentationMasksSAM,
)


def parse_detection_message(message: dai.ImgDetections):
    """Parses the detection message and returns the detections."""
    detections = message.detections
    return detections


def parse_line_detection_message(message: Lines):
    """Parses the line detection message and returns the lines."""
    lines = message.lines
    return lines


def parse_segmentation_message(message: dai.ImgFrame):
    """Parses the segmentation message and returns the mask."""
    mask = message.getFrame()
    mask = mask.reshape(mask.shape[0], mask.shape[1])
    return mask


def parse_keypoints_message(message: Keypoints):
    """Parses the keypoints message and returns the keypoints."""
    keypoints = message.keypoints
    return keypoints


def parse_classification_message(message: Classifications):
    """Parses the classification message and returns the classification."""
    classes = message.classes
    scores = message.scores
    return classes, scores


def parse_image_message(message: dai.ImgFrame):
    """Parses the image message and returns the image."""
    image = message.getFrame()
    return image


def parse_yolo_kpts_message(message: ImgDetectionsExtended):
    """Parses the yolo keypoints message and returns the keypoints."""
    detections = message.detections
    return detections


def parse_cluster_message(message: Clusters):
    """Parses the cluster message and returns the clusters."""
    clusters = message.clusters
    return clusters


def parse_fast_sam_message(message: SegmentationMasksSAM):
    """Parses the fast sam message and returns the masks."""
    masks = message.masks
    return masks


def parse_map_message(message: Map2D):
    """Parses the map message and returns the map."""
    map = message.map
    return map



import depthai as dai
from utils.arguments import initialize_argparser, parse_fps_limit, parse_model_slug
from utils.model import get_input_shape, get_model_from_hub, get_parser
from utils.xfeat import xfeat_mono, xfeat_stereo
from visualization.visualize import visualize

from depthai_nodes.parser_generator import ParserGenerator

# Initialize the argument parser
arg_parser, args = initialize_argparser()

# Parse the model slug
model_slug, model_version_slug = parse_model_slug(args)
fps_limit = parse_fps_limit(args)

# Get the model from the HubAI
nn_archive = get_model_from_hub(model_slug, model_version_slug)

# Get the parser
parser_class, parser_name = get_parser(nn_archive)
input_shape = get_input_shape(nn_archive)

if parser_name == "XFeatMonoParser":
    xfeat_mono(nn_archive, input_shape, fps_limit)
    exit(0)
elif parser_name == "XFeatStereoParser":
    xfeat_stereo(nn_archive, input_shape, fps_limit)
    exit(0)

# Create the pipeline
with dai.Pipeline() as pipeline:
    cam = pipeline.create(dai.node.Camera).build()

    # YOLO and MobileNet-SSD have native parsers in DAI - no need to create a separate parser
    if parser_name == "YOLO" or parser_name == "SSD":
        network = pipeline.create(dai.node.DetectionNetwork).build(
            cam.requestOutput(
                input_shape, type=dai.ImgFrame.Type.BGR888p, fps=fps_limit
            ),
            nn_archive,
        )
        parser_queue = network.out.createOutputQueue()
    else:
        image_type = dai.ImgFrame.Type.BGR888p
        if "gray" in model_version_slug:
            image_type = dai.ImgFrame.Type.GRAY8

        if input_shape[0] < 128 or input_shape[1] < 128:
            print(
                "Input shape is too small so we are requesting a larger image and resizing it."
            )
            print(
                "During visualization we resize small image back to large, so image quality is lower."
            )
            manip = pipeline.create(dai.node.ImageManip)
            manip.initialConfig.setResize(input_shape)
            large_input_shape = (input_shape[0] * 4, input_shape[1] * 4)
            cam.requestOutput(large_input_shape, type=image_type, fps=fps_limit).link(
                manip.inputImage
            )
            network = pipeline.create(dai.node.NeuralNetwork).build(
                manip.out, nn_archive
            )
        else:
            network = pipeline.create(dai.node.NeuralNetwork).build(
                cam.requestOutput(input_shape, type=image_type, fps=fps_limit),
                nn_archive,
            )

        parsers = pipeline.create(ParserGenerator).build(nn_archive)

        if len(parsers) == 0:
            raise ValueError("No parsers were generated.")

        if len(parsers) > 1:
            raise ValueError("Only models with one parser are supported.")

        parser = parsers[0]

        # Linking
        network.out.link(parser.input)

        parser_queue = parser.out.createOutputQueue()

    camera_queue = network.passthrough.createOutputQueue()

    pipeline.start()

    while pipeline.isRunning():
        frame: dai.ImgFrame = camera_queue.get().getCvFrame()
        message = parser_queue.get()

        extraParams = nn_archive.getConfig().model.heads[0].metadata.extraParams
        if visualize(frame, message, parser_name, extraParams):
            pipeline.stop()
            break
